/*
 * Author:      William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990-1993, William Cheng.
 *
 * Permission limited to the use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted by the Author without
 * fee, provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the Author not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.  All other
 * rights (including the right to sell "tgif" and the right to sell derivative
 * works of tgif) are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /amnt/maui/tangram/u/william/X11/TGIF2/RCS/file.c,v 2.226 1994/03/03 06:19:53 william Exp $";
#endif

#include <sys/types.h>
#include <sys/file.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "const.h"
#include "patchlevel.h"
#include "types.h"

#include "align.e"
#include "arc.e"
#include "attr.e"
#include "auxtext.e"
#include "box.e"
#include "button.e"
#include "choice.e"
#include "cmd.e"
#include "color.e"
#include "copypaste.e"
#include "cursor.e"
#include "dialog.e"
#include "drawing.e"
#include "dup.e"
#include "eps.e"
#ifndef _NO_EXTERN
#include "file.e"
#endif
#include "font.e"
#include "grid.e"
#include "group.e"
#include "mainloop.e"
#include "mainmenu.e"
#include "mark.e"
#include "menu.e"
#include "move.e"
#include "msg.e"
#include "names.e"
#include "obj.e"
#include "oval.e"
#include "page.e"
#include "pattern.e"
#include "poly.e"
#include "polygon.e"
#include "prtgif.e"
#include "ps.e"
#include "raster.e"
#include "rcbox.e"
#include "rect.e"
#include "ruler.e"
#include "scroll.e"
#include "select.e"
#include "setup.e"
#include "special.e"
#include "stk.e"
#include "stretch.e"
#include "text.e"
#include "version.e"
#include "xbitmap.e"
#include "xpixmap.e"

extern char	* mktemp ARGS_DECL((char *Template));
extern int	atoi ARGS_DECL((char *));
#ifndef _NO_EXTERN
extern time_t	time ARGS_DECL((time_t *));
extern int	unlink ARGS_DECL((char *));
#ifdef SYSV
extern unsigned	sleep ARGS_DECL((unsigned));
#else
extern int	sleep ARGS_DECL((unsigned));
#endif
#endif

#define CUR_VERSION 30
#define START_HAVING_ATTRS 8

extern char	* getenv ARGS_DECL((char *));
#ifdef __hpux
extern double	atof ARGS_DECL((const char *));
#else
extern double	atof ARGS_DECL((char *));
#endif

int	PRTGIF = FALSE;
char	curFileName[MAXPATHLENGTH+1];
int	curFileDefined = FALSE;
int	fileVersion = INVALID;
int	importingFile = FALSE;
int	psDotsPerInch = 72;
int	printMag = 100;
int	saveTmpOnReturn = TRUE;
int	warpToWinCenter = TRUE;
float	tiledPageScaling = 0.9;

#ifndef A4PAPER
char	*psXOffStr[MAXPAGESTYLES] = {"0","0"};
float	psXOff[MAXPAGESTYLES] =     { 0,  0 };
char	*psYOffStr[MAXPAGESTYLES] = {"11","0"};
float	psYOff[MAXPAGESTYLES] =     { 11,  0 };
float	psPageWidthInInch[MAXPAGESTYLES]  = { 8.5, 11 };
float	psPageHeightInInch[MAXPAGESTYLES] = { 11, 8.5 };
#else /* A4PAPER */
char	*psXOffStr[MAXPAGESTYLES] = {"0","0"};
float	psXOff[MAXPAGESTYLES] =     { 0,  0 };
char	*psYOffStr[MAXPAGESTYLES] = {"11.7","0"};
float	psYOff[MAXPAGESTYLES] =     { 11.7,  0 };
float	psPageWidthInInch[MAXPAGESTYLES]  = { 8.25,11.7 };
float	psPageHeightInInch[MAXPAGESTYLES] = { 11.7,8.25 };
#endif /* A4PAPER */

char	printCommand[MAXSTRING+1];
char	outputDir[MAXPATHLENGTH+1];

char	* savedComments=NULL;
int	savedCommentsLen=0;
int	saveCommentsInSaveNew=TRUE;
int	usePsAdobeString=FALSE;
char	adobeString[80], epsfString[80];

int	readingPageNum=0;
int	loadedCurPageNum=0;

int	writeFileFailed=FALSE;

struct DocFontRec {
   char name[80];
   int len;
   struct DocFontRec *next;
};

static struct DocFontRec *firstDocFont=NULL;

void UpdateDocumentFonts (ps_font_name)
   char	* ps_font_name;
{
   int			len=strlen(ps_font_name);
   struct DocFontRec	* df_ptr;

   for (df_ptr=firstDocFont; df_ptr!=NULL; df_ptr=df_ptr->next)
      if (df_ptr->len == len && strcmp(df_ptr->name,ps_font_name) == 0)
         return;
   df_ptr = (struct DocFontRec *) calloc (1, sizeof(struct DocFontRec));
   df_ptr->len = len;
   strcpy (df_ptr->name, ps_font_name);
   df_ptr->next = firstDocFont;
   firstDocFont = df_ptr;
}

static
int CopyAFile (file1, file2)
   char	* file1, * file2;
{
   char	tmp_str[MAXSTRING+1], * rest;
   int	short_name;
   FILE	* fp1, * fp2;

   if ((fp1 = fopen (file1, "r")) == NULL)
   {
      sprintf (tmp_str, "Can not open '%s' for read.", file1);
      if (PRTGIF)
         fprintf (stderr, "%s\n", tmp_str);
      else
         Msg (tmp_str);
      return (FALSE);
   }
   if ((short_name = IsPrefix (bootDir, file2, &rest))) ++rest;
   if ((fp2 = fopen (file2, "w")) == NULL)
   {
      if (PRTGIF)
         fprintf (stderr, "Can not open '%s' for write.\n", file2);
      else
      {
         if (short_name)
            sprintf (tmp_str, "Can not open '%s' for write.", rest);
         else
            sprintf (tmp_str, "Can not open '%s' for write.", file2);
         Msg (tmp_str);
      }
      fclose (fp1);
      return (FALSE);
   }
   writeFileFailed = FALSE;
   while (fgets (tmp_str, MAXSTRING, fp1) != NULL)
      if (fputs (tmp_str, fp2) == EOF)
         writeFileFailed = TRUE;
   fclose (fp1);
   fclose (fp2);
   if (writeFileFailed)
   {
      writeFileFailed = FALSE;
      sprintf (tmp_str, "Fail to write to '%s'.  File system may be full.",
            file2);
      if (PRTGIF)
         fprintf (stderr, "%s\n", tmp_str);
      else
         Dialog (tmp_str, NULL, NULL);
      return (FALSE);
   }
   return (TRUE);
}

static
int ExecuteCmd (cmd)
   char	* cmd;
{
   char	tmp_str[MAXSTRING+1];
   FILE	* fp;

   if ((fp = popen (cmd, "r")) == NULL) return (FALSE);

   while (fgets (tmp_str, MAXSTRING, fp) != NULL)
   {
      if (PRTGIF)
         fprintf (stderr, "%s", tmp_str);
      else
         Msg (tmp_str);
      sleep (5);
   }
   pclose (fp);
   return (TRUE);
}

void ClearFileInfo ()
{
   curFileName[0] = '\0';
   curFileDefined = FALSE;
   *curSymDir = '\0';
}

void CleanUpComments ()
{
   if (savedComments != NULL)
   {
      cfree (savedComments);
      savedComments = NULL;
      savedCommentsLen = 0;
   }
}

int OkayToCreateFile (FileName)
   char	* FileName;
{
   FILE	* fp;

   if ((fp = fopen (FileName, "r")) == NULL) return (TRUE);
   fclose (fp);
   switch (YesNoCancel ("File exists, okay to overwrite? [ync](y)",CONFIRM_YES))
   {
      case CONFIRM_YES: break;
      case CONFIRM_NO: return (FALSE);
      case CONFIRM_CANCEL: return (FALSE);
   }
   unlink (FileName);
   return (TRUE);
}

void Save (FP, BotObjPtr, Level, PageNumber)
   FILE			* FP;
   struct ObjRec	* BotObjPtr;
   int			Level, PageNumber;
{
   struct ObjRec	* obj_ptr;

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   if (Level == 0 && PageNumber == 1)
   {
      char	font_str[81];

      GetPSFontStr (curFont, curStyle, font_str);
      if (TGIF_PATCHLEVEL == 0)
      {
         if (fprintf (FP, "%%TGIF %s\n", version_string) == EOF)
            writeFileFailed = TRUE;
      }
      else
      {
         if (fprintf (FP, "%%TGIF %s-p%1d\n", version_string,
               TGIF_PATCHLEVEL) == EOF)
            writeFileFailed = TRUE;
      }

      if (fprintf (FP, "state(%1d,%1d,%1d,", pageStyle, CUR_VERSION,
            printMag) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,", drawOrigX, drawOrigY, zoomScale) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,", xyEnglishGrid, gridOn, colorIndex) ==
            EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,", horiAlign, vertAlign, lineWidth) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,%1d,", curSpline, lineStyle, objFill,
            penPat) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,'%s',%1d,%1d,", textJust,
            /* font_str starts with the '/' character */
            &font_str[1], curStyle, curSize) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,", 0, curDash, gridSystem) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,", xyMetricGrid, textVSpace, zoomedIn) ==
            EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,", gridShown, moveMode, curRotate) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,", rcbRadius, useGray) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%1d,%1d,%1d,%1d).\n", pageLayoutMode,
            (pageLayoutMode == PAGE_STACK) ? curPageNum : paperCol,
            (pageLayoutMode == PAGE_STACK) ? lastPageNum : paperRow,
            pageLineShownInTileMode) == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%%\n") == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%% @%s%s\n", "(#)$H", "eader$") == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%% %s\n", "%W%") == EOF)
         writeFileFailed = TRUE;
      if (fprintf (FP, "%%\n") == EOF)
         writeFileFailed = TRUE;

      if (savedComments != NULL)
         if (fputs (savedComments, FP) == EOF)
            writeFileFailed = TRUE;

      if (tgifObj->lattr != NULL)
      {
         if (fprintf (FP, "file_attr(") == EOF) writeFileFailed = TRUE;
         SaveAttrs (FP, tgifObj->lattr);
         if (fprintf (FP, ").\n") == EOF) writeFileFailed = TRUE;
      }
   }
   if (Level == 0)
      if (fprintf (FP, "page(%1d,\"%s\").\n", PageNumber,
            ((pageLayoutMode==PAGE_TILE || curPage->name==NULL) ? "" :
            curPage->name)) == EOF)
         writeFileFailed = TRUE;

   for (obj_ptr = BotObjPtr; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
   {
      switch (obj_ptr->type)
      {
         case OBJ_POLY: SavePolyObj (FP, obj_ptr); break;
         case OBJ_BOX: SaveBoxObj (FP, obj_ptr); break;
         case OBJ_OVAL: SaveOvalObj (FP, obj_ptr); break;
         case OBJ_TEXT: SaveTextObj (FP, obj_ptr); break;
         case OBJ_POLYGON: SavePolygonObj (FP, obj_ptr); break;
         case OBJ_ARC: SaveArcObj (FP, obj_ptr); break;
         case OBJ_RCBOX: SaveRCBoxObj (FP, obj_ptr); break;
         case OBJ_XBM: SaveXBmObj (FP, obj_ptr); break;
         case OBJ_XPM: SaveXPmObj (FP, obj_ptr); break;
         case OBJ_GROUP: SaveGroupObj (FP, obj_ptr, Level); break;
         case OBJ_SYM: SaveCompObj (FP, obj_ptr, Level); break;
         case OBJ_ICON: SaveIconObj (FP, obj_ptr, Level); break;
      }
      if (obj_ptr->prev == NULL)
      {
         if (Level == 0)
         {
            if (fprintf (FP, ".\n") == EOF) writeFileFailed = TRUE;
         }
         else
         {
            if (fprintf (FP, "\n") == EOF) writeFileFailed = TRUE;
         }
      }
      else
      {
         if (Level == 0)
         {
            if (fprintf (FP, ".\n") == EOF) writeFileFailed = TRUE;
         }
         else
         {
            if (fprintf (FP, ",\n") == EOF) writeFileFailed = TRUE;
         }
      }
   }
   SetCurChoice (NOTHING);
   SetDefaultCursor (mainWindow);
   ShowCursor ();
}

int SaveTmpFile (NewFileName)
   char	* NewFileName;
   /* return TRUE if file successfully saved */
{
   char			new_file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1];
   char			* rest;
   FILE			* fp;
   int			count = 0, status, short_name;
   struct PageRec	* saved_cur_page;
   struct ObjRec	* obj_ptr;

   strcpy (new_file_name, NewFileName);

   saved_cur_page = curPage;
   for (curPage=firstPage; curPage != NULL; curPage = curPage->next)
      for (obj_ptr = curPage->top; obj_ptr != NULL; obj_ptr = obj_ptr->next)
         if (obj_ptr->type == OBJ_SYM)
            count++;
   curPage = saved_cur_page;

   switch (count)
   {
      case 0:
         sprintf (new_file_name, "%s.%s", NewFileName, OBJ_FILE_EXT);
         status = OBJ_FILE_SAVED;
         break;
      case 1:
         if (lastPageNum == 1)
         {
            sprintf (new_file_name, "%s.%s", NewFileName, SYM_FILE_EXT);
            status = SYM_FILE_SAVED;
         }
         else
         {
            Dialog ("A symbol file can only have one page.  Save aborted!",
                  "( <CR> or <ESC> to continue )", new_file_name);
            return (INVALID);
         }
         break;
      default:
         Msg ("TOO MANY SYMBOLS!  Symbol file not saved.");
         return (INVALID);
   }

   unlink (new_file_name);

   if ((short_name = IsPrefix (bootDir, new_file_name, &rest))) ++rest;
   if ((fp = fopen (new_file_name, "w")) == NULL)
   {
      if (short_name)
         sprintf (s, "Can not open '%s', file not saved.", rest);
      else
         sprintf (s, "Can not open '%s', file not saved.", new_file_name);
      Msg (s);
      return (INVALID);
   }

   if (short_name)
      sprintf (s, "Saving temporary file '%s' ...", rest);
   else
      sprintf (s, "Saving temporary file '%s' ...", new_file_name);
   Msg (s);

   writeFileFailed = FALSE;
   MakeQuiescent ();
   saved_cur_page = curPage;
   for (curPage = firstPage, count = 1; curPage != NULL;
         curPage = curPage->next, count++)
   {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save (fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose (fp);

   if (writeFileFailed)
   {
      writeFileFailed = FALSE;
      sprintf (s, "Fail to write to '%s'.  File system may be full.",
            new_file_name);
      Dialog (s, NULL, NULL);
   }
   else
   {
      if (short_name)
         sprintf (s, "Temporary file '%s' saved.", rest);
      else
         sprintf (s, "Temporary file '%s' saved.", new_file_name);
      Msg (s);
   }

   if (PSFILE_MOD != 0 && chmod (new_file_name, 0777))
   {
      if (short_name)
         sprintf (s, "Can not chmod '%s' to 0777.", rest);
      else
         sprintf (s, "Can not chmod '%s' to 0777.", new_file_name);
      Msg (s);
   }
   return (status);
}

void SaveNewFile (SaveSelectedOnly)
   int	SaveSelectedOnly;
{
   char			new_file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1];
   char			new_full_name[MAXPATHLENGTH+1];
   char			tmp_str[MAXPATHLENGTH+1];
   char			name_without_ext[MAXPATHLENGTH+1], * rest;
   char			saved_cur_dir[MAXPATHLENGTH+1];
   char			saved_cur_file_name[MAXPATHLENGTH+1];
   char			obj_ext_str[MAXSTRING+1], sym_ext_str[MAXSTRING+1];
   int			saved_cur_file_defined=FALSE, obj_ext_len, sym_ext_len;
   FILE			* fp;
   int			count = 0, len, short_name, ok=TRUE;
   struct ObjRec	* obj_ptr, * saved_top_obj=NULL, * saved_bot_obj=NULL;
   struct SelRec	* top_sel_ptr, * bot_sel_ptr;
   struct SelRec	* sel_ptr, * next_sel;
   struct PageRec	* saved_cur_page;

   if (SaveSelectedOnly && topSel==NULL)
   {
      Msg ("No objects selected!  Nothing saved!");
      return;
   }

   Dialog ("Please enter new file name:", "( <CR>: accept, <ESC>: cancel )",
         new_file_name);
   if (*new_file_name == '\0') return;
   len = strlen (new_file_name);

   if (SaveSelectedOnly)
   {
      for (sel_ptr = topSel; sel_ptr != NULL; sel_ptr = sel_ptr->next)
         if (sel_ptr->obj->type == OBJ_SYM)
            count++;
   }
   else
   {
      saved_cur_page = curPage;
      for (curPage = firstPage; curPage != NULL; curPage = curPage->next)
         for (obj_ptr = curPage->top; obj_ptr != NULL; obj_ptr = obj_ptr->next)
            if (obj_ptr->type == OBJ_SYM)
               count++;
      curPage = saved_cur_page;
   }

   sprintf (obj_ext_str, ".%s", OBJ_FILE_EXT);
   sprintf (sym_ext_str, ".%s", SYM_FILE_EXT);
   obj_ext_len = strlen (obj_ext_str);
   sym_ext_len = strlen (sym_ext_str);

   switch (count)
   {
      case 0:
         if (len >= obj_ext_len || len >= sym_ext_len)
         {
            if (strcmp (&new_file_name[len-sym_ext_len], sym_ext_str) == 0)
            {
               Msg ("Can not save as a symbol file, no symbol defined.");
               return;
            }
            else if (strcmp (&new_file_name[len-obj_ext_len], obj_ext_str) != 0)
            {
               strcpy (name_without_ext, new_file_name);
               strcat (new_file_name, obj_ext_str);
            }
            else
            {
               strcpy (name_without_ext, new_file_name);
               name_without_ext[len-obj_ext_len] = '\0';
            }

            if (strlen (new_file_name) == obj_ext_len)
            {
               Msg ("No file name specified.  File not saved.");
               return;
            }
         }
         else
         {
            strcpy (name_without_ext, new_file_name);
            strcat (new_file_name, obj_ext_str);
         }
         break;
      case 1:
         if (lastPageNum == 1)
         {
            if (len >= obj_ext_len || len >= sym_ext_len)
            {
               if (strcmp (&new_file_name[len-obj_ext_len], obj_ext_str) == 0)
               {
                  Msg ("Can not save as an object file; one symbol defined.");
                  return;
               }
               else if (strcmp(&new_file_name[len-sym_ext_len],sym_ext_str)!=0)
               {
                  strcpy (name_without_ext, new_file_name);
                  strcat (new_file_name, sym_ext_str);
               }
               else
               {
                  strcpy (name_without_ext, new_file_name);
                  name_without_ext[len-sym_ext_len] = '\0';
               }

               if (strlen (new_file_name) == sym_ext_len)
               {
                  Msg ("No file name specified.  File not saved.");
                  return;
               }
            }
            else
            {
               strcpy (name_without_ext, new_file_name);
               strcat (new_file_name, sym_ext_str);
            }
         }
         else
         {
            Dialog ("A symbol file can only have one page.  Save aborted!",
                  "( <CR> or <ESC> to continue )", new_file_name);
            return;
         }
         break;
      default:
         Msg ("TOO MANY SYMBOLS!  Symbol file not saved.");
         return;
   }

   if (*new_file_name == '/')
      strcpy (new_full_name, new_file_name);
   else
      sprintf (new_full_name, "%s/%s", curDir, new_file_name);

   if (!OkayToCreateFile (new_full_name)) return;

   if ((short_name = IsPrefix (bootDir, new_full_name, &rest))) ++rest;
   if ((fp = fopen (new_full_name, "w")) == NULL)
   {
      if (short_name)
         sprintf (s, "Can not open '%s', file not saved.", rest);
      else
         sprintf (s, "Can not open '%s', file not saved.", new_full_name);
      Msg (s);
      return;
   }

   if (SaveSelectedOnly)
   {
      PushPageInfo ();
      if (pageLayoutMode == PAGE_STACK)
         paperCol = paperRow = curPageNum = lastPageNum = 1;

      strcpy (saved_cur_dir, curDir);
      strcpy (saved_cur_file_name, curFileName);
      saved_cur_file_defined = curFileDefined;
      saved_top_obj = topObj;
      saved_bot_obj = botObj;

      JustDupSelObj (&top_sel_ptr, &bot_sel_ptr);
      topObj = top_sel_ptr->obj;
      botObj = bot_sel_ptr->obj;

      firstPage = lastPage = curPage =
            (struct PageRec *) calloc (1, sizeof(struct PageRec));
      firstPage->top = topObj;
      firstPage->bot = botObj;
      firstPage->next = firstPage->prev = NULL;
      if (pageLayoutMode == PAGE_STACK)
      {
         firstPage->draw_orig_x = drawOrigX;
         firstPage->draw_orig_y = drawOrigY;
         firstPage->zoom_scale = zoomScale;
         firstPage->zoomed_in = zoomedIn;
         curPageNum = lastPageNum = 1;
      }

      for (sel_ptr=topSel, obj_ptr=topObj; obj_ptr!=NULL;
            sel_ptr=sel_ptr->next, obj_ptr=obj_ptr->next)
      {
         CopyObjId (sel_ptr->obj, obj_ptr);
         CopyObjLocks (sel_ptr->obj, obj_ptr);
      }
   }

   strcpy (tmp_str, curDir);
   SetCurDir (new_full_name);
   curFileDefined = TRUE;

   switch (count)
   {
      case 0:
         *curSymDir = '\0';
         if ((strcmp (tmp_str, curDir) != 0) || (!NameInCurDir (curFileName)))
            UpdateDirInfo ();
         break;
      case 1:
         strcpy (curSymDir, curDir);
         if (!DirInSymPath (curDir)) UpdateSymInfo ();
         break;
   }

   if (short_name)
      sprintf (s, "Saving '%s' ...", rest);
   else
      sprintf (s, "Saving '%s' ...", new_full_name);
   Msg (s);

   if (!saveCommentsInSaveNew) CleanUpComments ();
   writeFileFailed = FALSE;
   MakeQuiescent ();
   saved_cur_page = curPage;
   for (curPage = firstPage, count = 1; curPage != NULL;
         curPage = curPage->next, count++)
   {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save (fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose (fp);

   if (writeFileFailed)
   {
      writeFileFailed = FALSE;
      sprintf (s, "Fail to write to '%s'.  File system may be full.",
            new_full_name);
      Dialog (s, NULL, NULL);
      ok = FALSE;
   }
   else
   {
      if (short_name)
         sprintf (s, "File '%s' saved.", rest);
      else
         sprintf (s, "File '%s' saved.", new_full_name);
      Msg (s);
   }

   if (SaveSelectedOnly)
   {
      CleanUpPage ();
      for (sel_ptr = top_sel_ptr; sel_ptr != NULL; sel_ptr = next_sel)
      {
         next_sel = sel_ptr->next;
         cfree (sel_ptr);
      }

      topObj = saved_top_obj;
      botObj = saved_bot_obj;
      strcpy (curDir, saved_cur_dir);
      strcpy (curFileName, saved_cur_file_name);
      curFileDefined = saved_cur_file_defined;
      PopPageInfo ();
   }
   else if (ok)
   {
      SetFileModified (FALSE);
      RedrawTitleWindow ();
   }
}

void SaveSymInLibrary ()
{
   char			new_file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1];
   char			new_full_name[MAXPATHLENGTH+1];
   char			dir_name[MAXPATHLENGTH+1];
   char			saved_dir[MAXPATHLENGTH+1], saved_file[MAXPATHLENGTH+1];
   char			saved_sym_dir[MAXPATHLENGTH+1];
   char			name_without_ext[MAXPATHLENGTH+1], * rest;
   char			sym_ext_str[MAXSTRING+1], * c_ptr;
   FILE			* fp;
   int			count=0, len, short_name, sym_ext_len;
   int			saved_cur_file_defined;
   struct ObjRec	* obj_ptr;
   struct PageRec	* saved_cur_page;

   saved_cur_page = curPage;
   for (curPage = firstPage; curPage != NULL; curPage = curPage->next)
      for (obj_ptr = curPage->top; obj_ptr != NULL; obj_ptr = obj_ptr->next)
         if (obj_ptr->type == OBJ_SYM)
            count++;
   curPage = saved_cur_page;

   switch (count)
   {
      case 0: Msg ("No Symbol found!  Symbol file not saved."); return;
      case 1:
         if (lastPageNum != 1)
         {
            Dialog ("A symbol file can only have one page.  Save aborted!",
                  "( <CR> or <ESC> to continue )", new_file_name);
            return;
         }
         break;
      default: Msg ("TOO MANY SYMBOLS!  Symbol file not saved."); return;
   }

   sprintf (sym_ext_str, ".%s", SYM_FILE_EXT);
   sym_ext_len = strlen (sym_ext_str);

   if (*curFileName == '\0')
   {
      Dialog ("Please enter new file name:", "( <CR>: accept, <ESC>: cancel )",
            new_file_name);
      if (*new_file_name == '\0') return;

      len = strlen (new_file_name);
      if (len >= sym_ext_len)
      {
         if (strcmp (&new_file_name[len-sym_ext_len], sym_ext_str) != 0)
         {
            strcpy (name_without_ext, new_file_name);
            strcat (new_file_name, sym_ext_str);
         }
         else
         {
            strcpy (name_without_ext, new_file_name);
            name_without_ext[len-sym_ext_len] = '\0';
         }
   
         if (strlen (new_file_name) == sym_ext_len)
         {
            Msg ("No file name specified.  File not saved.");
            return;
         }
      }
      else
      {
         strcpy (name_without_ext, new_file_name);
         strcat (new_file_name, sym_ext_str);
      }
   }
   else
   {
      len = strlen (curFileName);
      for (c_ptr = &curFileName[len-1]; c_ptr != curFileName; c_ptr--)
         if (*c_ptr == '/')
            break;
      if (*c_ptr == '/') c_ptr++;
      strcpy (new_file_name, c_ptr);
   }

   if (SelectSymDir (dir_name) == INVALID) { Msg (""); return; }

   if (strcmp (dir_name, ".") == 0)
      sprintf (new_full_name, "%s/%s", curDir, new_file_name);
   else
      sprintf (new_full_name, "%s/%s", dir_name, new_file_name);

   if (!OkayToCreateFile (new_full_name)) return;

   if ((short_name = IsPrefix (bootDir, new_full_name, &rest))) ++rest;
   if ((fp = fopen (new_full_name, "w")) == NULL)
   {
      if (short_name)
         sprintf (s, "Can not open '%s', file not saved.", rest);
      else
         sprintf (s, "Can not open '%s', file not saved.", new_full_name);
      Msg (s);
      return;
   }

   strcpy (saved_dir, curDir);
   strcpy (saved_file, curFileName);
   strcpy (saved_sym_dir, curSymDir);
   saved_cur_file_defined = curFileDefined;
   SetCurDir (new_full_name);
   curFileDefined = TRUE;

   strcpy (curSymDir, curDir);
   if (!DirInSymPath (curDir)) UpdateSymInfo ();

   if (short_name)
      sprintf (s, "Saving '%s' ...", rest);
   else
      sprintf (s, "Saving '%s' ...", new_full_name);
   Msg (s);

   if (!saveCommentsInSaveNew) CleanUpComments ();
   writeFileFailed = FALSE;
   MakeQuiescent ();
   saved_cur_page = curPage;
   for (curPage = firstPage, count = 1; curPage != NULL;
         curPage = curPage->next, count++)
   {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save (fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose (fp);

   if (writeFileFailed)
   {
      writeFileFailed = FALSE;
      sprintf (s, "Fail to write to '%s'.  File system may be full.",
            new_full_name);
      Dialog (s, NULL, NULL);
   }
   else
   {
      if (short_name)
         sprintf (s, "File '%s' saved.", rest);
      else
         sprintf (s, "File '%s' saved.", new_full_name);
      Msg (s);
   }

   strcpy (curDir, saved_dir);
   strcpy (curFileName, saved_file);
   strcpy (curSymDir, saved_sym_dir);
   curFileDefined = saved_cur_file_defined;

   RedrawTitleWindow ();
}

void SaveFile ()
{
   int			i, len, count=0, short_name;
   struct ObjRec	* obj_ptr;
   FILE			* fp;
   char			ext[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1];
   char			full_name[MAXPATHLENGTH+1], * rest;
   struct PageRec	* saved_cur_page;

   if (!curFileDefined)
   {
      SaveNewFile (FALSE);
      return;
   }

   len = strlen (curFileName);
   for (i = len-1; curFileName[i] != '.'; i--) ;
   strcpy (ext, &curFileName[i+1]);

   saved_cur_page = curPage;
   for (curPage = firstPage; curPage != NULL; curPage = curPage->next)
      for (obj_ptr = curPage->top; obj_ptr != NULL; obj_ptr = obj_ptr->next)
         if (obj_ptr->type == OBJ_SYM)
            count++;
   curPage = saved_cur_page;

   switch (count)
   {
      case 0:
         if (strcmp (ext, SYM_FILE_EXT) == 0)
         {
            Msg ("No SYMBOL defined in symbol file.  Symbol file not saved.");
            return;
         }
         break;
      case 1:
         if (strcmp (ext, OBJ_FILE_EXT) == 0)
         {
            Msg ("One SYMBOL defined in OBJECT file.  Object file not saved.");
            return;
         }
         break;
      default:
         if (strcmp (ext, SYM_FILE_EXT) == 0)
         {
            Msg ("Too many SYMBOLS in symbol file!  Symbol file not saved.");
            return;
         }
         break;
   }

   if (strcmp (ext, SYM_FILE_EXT) == 0)
      sprintf (full_name, "%s/%s", curSymDir, curFileName);
   else if (strcmp (ext, OBJ_FILE_EXT) == 0)
      sprintf (full_name, "%s/%s", curDir, curFileName);

   if ((short_name = IsPrefix (bootDir, full_name, &rest))) ++rest;
   if ((fp = fopen (full_name, "w")) == NULL)
   {
      if (short_name)
         sprintf (s, "Can not open '%s', file not saved.", rest);
      else
         sprintf (s, "Can not open '%s', file not saved.", full_name);
      Msg (s);
      return;
   }

   if (short_name)
      sprintf (s, "Saving '%s' ...", rest);
   else
      sprintf (s, "Saving '%s' ...", full_name);
   Msg (s);

   writeFileFailed = FALSE;
   MakeQuiescent ();
   saved_cur_page = curPage;
   for (curPage = firstPage, count = 1; curPage != NULL;
         curPage = curPage->next, count++)
   {
      topObj = curPage->top;
      botObj = curPage->bot;
      Save (fp, botObj, 0, count);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;
   fclose (fp);

   if (writeFileFailed)
   {
      writeFileFailed = FALSE;
      sprintf (s, "Fail to write to '%s'.  File system may be full.",
            full_name);
      Dialog (s, NULL, NULL);
   }
   else
   {
      if (short_name)
         sprintf (s, "File '%s' saved.", rest);
      else
         sprintf (s, "File '%s' saved.", full_name);
      Msg (s);
      SetFileModified (FALSE);
      if (!NameInCurDir (curFileName)) UpdateDirInfo ();
   }
}

char * ParseStr (Str, C, Left)
   char	* Str, C, * Left;
{
   register char	* s = Str, * l = Left;
   register int		len = 0;

   while (*s != '\0' && *s != C)
   {
      *l++ = *s++;
      len++;
   }

   if (*s == C) s++;
   *l = '\0';

   while (len >= 2 && *Left == '\'' && *(l-1) == '\'')
   {
      *(--l) = '\0';
      strcpy (Left, &Left[1]);
      len -= 2;
   }
   return (s);
}

char * FindChar (C, Str)
   char	C;
   char	* Str;
   /* returns the address of the character right after C of the string Str */
{
   register char	* s = Str;

   while (*s != '\0' && *s != C) s++;

   if (*s == C) s++;
   return (s);
}

#define GETVALUE(val,name) \
      if (ScanValue("%d", &(val), name, "state") == INVALID) return (FALSE)
#define GETSTRNG(val,name) \
      if (ScanValue("%s", (val), name, "state") == INVALID) return (FALSE)

static
int ReadState (Inbuf)
   char	* Inbuf;
{
   char	* s, font_str[81];
   int	page_style, forced_use_gray=FALSE, compat_dpi;

   s = FindChar ('(', Inbuf);
   if (sscanf (s, "%d", &page_style) != 1) return (FALSE);
   s = FindChar (',', s);
   if (*s == '\0')
      fileVersion = INVALID;
   else if (sscanf (s, "%d", &fileVersion) != 1)
      return (FALSE);

   if (fileVersion > CUR_VERSION) return (FALSE);

   if (!importingFile)
   {
      char msg[MAXSTRING];

      if (fileVersion <= 13)
      {
         switch (page_style)
         {
            case PORTRAIT: printMag = 100; break;
            case LANDSCAPE: printMag = 100; break;
            case HIGHPORT: printMag = 50; page_style = PORTRAIT; break;
            case HIGHLAND: printMag = 50; page_style = LANDSCAPE; break;
            case SLIDEPORT: printMag = 200; page_style = PORTRAIT; break;
            case SLIDELAND: printMag = 200; page_style = LANDSCAPE; break;
            default:
               sprintf (msg, "Unrecognizable page style '%1d'.", page_style);
               TwoLineMsg (msg, "Portrait style assumed.");
               page_style = PORTRAIT;
               printMag = 100;
               break;
         }
      }
      else
      {
         if (page_style != PORTRAIT && page_style != LANDSCAPE)
         {
            sprintf (msg, "Unrecognizable page style '%1d'.", page_style);
            TwoLineMsg (msg, "Portrait style assumed.");
            page_style = PORTRAIT;
         }
         s = FindChar(',', s);
         sscanf (s, "%d", &printMag);
      }
      pageStyle = page_style;
   }

   if (PRTGIF && useGray) forced_use_gray = TRUE;
   if (!importingFile)
   {
      int page_arg1=1, page_arg2=1;

      if (fileVersion >= 2)
      {
         compat_dpi = FONT_DPI_75;
         curDash = 0;
         gridSystem = ENGLISH_GRID;
         xyMetricGrid = DEFAULT_METRIC_GRID;
         textVSpace = 0;
         zoomedIn = FALSE;
         curRotate = ROTATE0;
         rcbRadius = DEF_RCB_RADIUS;
         pageLayoutMode = PAGE_STACK;
         paperCol = paperRow = 1;
         curPageNum = lastPageNum = 1;
         pageLineShownInTileMode = TRUE;

         s = FindChar (',', s);

         InitScan (s, "\t\n, ");

         if (fileVersion <= 3)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");

            if (lineWidth == LINE_CURVED)
            {
               lineWidth = 0;
               curSpline = LT_SPLINE;
            }
            else
               curSpline = LT_STRAIGHT;
         }
         else if (fileVersion <= 7)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
         }
         else if (fileVersion <= 8)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
         }
         else if (fileVersion <= 11)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
         }
         else if (fileVersion <= 12)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
         }
         else if (fileVersion <= 18)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
         }
         else if (fileVersion <= 19)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
         }
         else if (fileVersion <= 21)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
            GETVALUE(gridShown,     "Grid Shown");
            GETVALUE(moveMode,      "Move Mode");
         }
         else if (fileVersion <= 26)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
            GETVALUE(gridShown,     "Grid Shown");
            GETVALUE(moveMode,      "Move Mode");
            GETVALUE(curRotate,     "Text Rotation");
            GETVALUE(rcbRadius,     "RCBox Radius");
         }
         else if (fileVersion <= 27)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
            GETVALUE(gridShown,     "Grid Shown");
            GETVALUE(moveMode,      "Move Mode");
            GETVALUE(curRotate,     "Text Rotation");
            GETVALUE(rcbRadius,     "RCBox Radius");
            GETVALUE(useGray,       "Use Gray Scale");
         }
         else if (fileVersion <= 28)
         {  /* Matsuda's Version */
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
            GETVALUE(gridShown,     "Grid Shown");
            GETVALUE(moveMode,      "Move Mode");
            GETVALUE(curRotate,     "Text Rotation");
            GETVALUE(rcbRadius,     "RCBox Radius");
            GETVALUE(useGray,       "Use Gray Scale");
            GETVALUE(paperCol,      "Num Columns");
            GETVALUE(paperRow,      "Num Rows");
            GETVALUE(curPageNum,    "Current Page Number");
            GETVALUE(lastPageNum,   "Last Page Number");
         }
         else if (fileVersion <= 29)
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETVALUE(curFont,       "Font Name");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
            GETVALUE(gridShown,     "Grid Shown");
            GETVALUE(moveMode,      "Move Mode");
            GETVALUE(curRotate,     "Text Rotation");
            GETVALUE(rcbRadius,     "RCBox Radius");
            GETVALUE(useGray,       "Use Gray Scale");
            GETVALUE(pageLayoutMode,"Page Layout Mode");
            GETVALUE(page_arg1,     "Page Layout Subarg 1");
            GETVALUE(page_arg2,     "Page Layout Subarg 2");
            GETVALUE(pageLineShownInTileMode,"Page Lines Shown");
         }
         else
         {
            GETVALUE(drawOrigX,     "X Draw Origin");
            GETVALUE(drawOrigY,     "Y Draw Origin");
            GETVALUE(zoomScale,     "Zoom scale");
            GETVALUE(xyEnglishGrid, "English Grid");
            GETVALUE(gridOn,        "Grid");
            GETVALUE(colorIndex,    "Color");
            GETVALUE(horiAlign,     "Horizontal Align");
            GETVALUE(vertAlign,     "Vertical Align");
            GETVALUE(lineWidth,     "Line Width");
            GETVALUE(curSpline,     "Spline");
            GETVALUE(lineStyle,     "Line Style");
            GETVALUE(objFill,       "Fill Pattern");
            GETVALUE(penPat,        "Pen Pattern");
            GETVALUE(textJust,      "Text Justify");
            GETSTRNG(font_str,      "Font Name String");
            GETVALUE(curStyle,      "Font Style");
            GETVALUE(curSize,       "Font Size");
            GETVALUE(compat_dpi,    "Font DPI");
            GETVALUE(curDash,       "Dash Style");
            GETVALUE(gridSystem,    "Grid System");
            GETVALUE(xyMetricGrid,  "Metric Grid");
            GETVALUE(textVSpace,    "Text Vertical Spacing");
            GETVALUE(zoomedIn,      "Zoomed In");
            GETVALUE(gridShown,     "Grid Shown");
            GETVALUE(moveMode,      "Move Mode");
            GETVALUE(curRotate,     "Text Rotation");
            GETVALUE(rcbRadius,     "RCBox Radius");
            GETVALUE(useGray,       "Use Gray Scale");
            GETVALUE(pageLayoutMode,"Page Layout Mode");
            GETVALUE(page_arg1,     "Page Layout Subarg 1");
            GETVALUE(page_arg2,     "Page Layout Subarg 2");
            GETVALUE(pageLineShownInTileMode,"Page Lines Shown");
         }
         if (fileVersion <= 28) readingPageNum++;
         if (fileVersion <= 29)
            curSize = GetCompatibleSize(compat_dpi,curSize);
         else
         {
            int		len=strlen(font_str);
            char	*s, msg[MAXSTRING];

            if (*font_str == '\'' && font_str[len-1] == '\'')
            {
               font_str[len-1] = '\0';
               s = &font_str[1];
            }
            else
               s = font_str;
            curFont = GetFontIndex (s, curStyle, TRUE);
            if (curFont == INVALID && !PRTGIF)
            {
               sprintf (msg, "Can not find screen font for '%s'.", s);
               TwoLineMsg (msg, "    Use Times instead.");
               curFont = FONT_TIM;
            }
         }
         switch (pageLayoutMode)
         {
            case PAGE_STACK:
               curPageNum = page_arg1;
               lastPageNum = page_arg2;
               paperCol = paperRow = 1;
               pageLineShownInTileMode = TRUE;
               break;
            case PAGE_TILE:
               paperCol = page_arg1;
               paperRow = page_arg2;
               curPageNum = lastPageNum = 1;
               break;
         }
         loadedCurPageNum = curPageNum;
         InitPage ();
         curPageNum = loadedCurPageNum;

         if (PRTGIF)
         {
            if (forced_use_gray) useGray = TRUE;
            return (TRUE);
         }
         if (colorIndex >= maxColors)
         {
            fprintf (stderr, "%s #%1d, use '%s' %s.\n",
                  "In reading state, can not find color", colorIndex,
                  colorMenuItems[defaultColorIndex], "as the current color");
            colorIndex = defaultColorIndex;
         }
         SetCanvasFont ();
         if (fileVersion <= 13)
         {
            switch (gridSystem)
            {
               case ENGLISH_GRID:
                  drawOrigX += HALF_INCH;
                  drawOrigY += HALF_INCH;
                  break;
               case METRIC_GRID:
                  drawOrigX += 2.5*ONE_CM;
                  drawOrigY += 2.5*ONE_CM;
                  break;
            }
         }
      }
      UpdPageStyle (pageStyle);
      if (PRTGIF) return (TRUE);

      if (lineWidth >= maxLineWidths)
      {
         fprintf (stderr, "%s '%1d' is out of range!  Set to 0.\n",
               "File's linewidth index", lineWidth);
         lineWidth = 0;
      }

      UpdDrawWinWH ();
      RedrawScrollBars ();
      ShowPage ();
      ShowPageLayout ();
      UpdDrawWinBBox ();

      SetDefaultDrawWinClipRecs ();

      DrawPaperBoundary ();
      RedrawGridLines ();
      RedrawPageLines ();
      RedrawRulers ();
      RedrawChoiceWindow ();
   }
   return (TRUE);
}

static
void ReadObjAttrs (MinFileVersion, FP, ObjPtr)
   int			MinFileVersion;
   FILE			* FP;
   struct ObjRec	* * ObjPtr;
{
   struct AttrRec	* top_attr = NULL, * bot_attr = NULL, * attr_ptr;

   if (fileVersion <= MinFileVersion) return;

   while (ReadAttr (FP, &attr_ptr))
   {
      attr_ptr->owner = *ObjPtr;
      attr_ptr->prev = NULL;
      attr_ptr->next = top_attr;
      if (top_attr == NULL)
         bot_attr = attr_ptr;
      else
         top_attr->prev = attr_ptr;
      top_attr = attr_ptr;
   }
   if (bot_attr != NULL) bot_attr->next = NULL;
   (*ObjPtr)->fattr = top_attr;
   (*ObjPtr)->lattr = bot_attr;
}

static
void ReadPageObj (Inbuf)
   char	* Inbuf;
{
   int			page_num;
   char			* s, * c_ptr, dummy_str[MAXSTRING+1], msg[MAXSTRING+1];

   s = FindChar ('(', Inbuf);
   if (sscanf (s, "%d", &page_num) != 1)
   {
      (void) sprintf(msg, "%s, %d:  Missing %s in %s.",
            scanFileName, scanLineNum, "page_num", "page");
      if (PRTGIF)
         fprintf (stderr, "%s\n", msg);
      else
         Msg (msg);
      return;
   }
   if (!importingFile)
   {
      for (curPage = firstPage; page_num != 1 && curPage != NULL;
            curPage = curPage->next, page_num--) ;
      if (curPage == NULL)
      {
         sprintf (msg, "Malformed input file (%s %1d).  Abort!",
               "apparently reading page", page_num);
         if (PRTGIF)
            fprintf (stderr, "%s\n", msg);
         else
            Dialog (msg,  "( <CR> or <ESC> to continue and abort )", dummy_str);
         exit (-1);
      }
   }
   s = FindChar (',', s);
   c_ptr = FindChar ('"', s);
   s = ReadString (c_ptr);
   *(--s) = '\0';
   if (*c_ptr != '\0')
   {
      curPage->name = (char *) calloc (strlen (c_ptr) + 1, sizeof(char));
      strcpy (curPage->name, c_ptr);
   }
   if (!importingFile)
   {
      topObj = curPage->top;
      botObj = curPage->bot;
   }
   readingPageNum++;
}

int ReadObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* * ObjPtr;
{
   char			inbuf[MAXSTRING+1], obj_name[10];
   int			len, cur_size, done, allocated, read_state_ok;

   *ObjPtr = NULL;
   while (fgets (inbuf, MAXSTRING, FP) != NULL)
   {
      char	* line, * c_ptr;

      scanLineNum++;
      if (inbuf[0] == ']') return (FALSE);
      allocated = FALSE;

      len = strlen(inbuf);
      if (inbuf[len-1] != '\r' && inbuf[len-1] != '\n')
      {  /* line longer than MAXSTRING characters */
         /* inbuf[MAXSTRING-1] == '\0' and len == MAXSTRING-1 now */
         cur_size = 2*MAXSTRING-1;
         line = (char *) calloc (cur_size+1, sizeof(char));
         strcpy (line, inbuf);
         c_ptr = &(line[MAXSTRING-1]);
         allocated = TRUE;

         done = FALSE;
         while (!done && fgets (inbuf, MAXSTRING, FP) != NULL)
         {
            scanLineNum++;
            len = strlen(inbuf);
            if (inbuf[len-1] == '\r' || inbuf[len-1] == '\n')
            {
               done = TRUE;
               inbuf[len-1] = '\0';
               strcpy (c_ptr, inbuf);
            }
            else
            {
               int	n = c_ptr - line;	/* NMH */

               cur_size += MAXSTRING-1;
               line = (char *) realloc (line, cur_size+1);
               c_ptr = line + n;		/* NMH */
               strcpy (c_ptr, inbuf);
               c_ptr += MAXSTRING-1;
            }
         }
         if (!done)
         {
            sprintf (inbuf, "%s, %d:  EOF in ReadObj ().  Read aborted!",
                  scanFileName, scanLineNum);
            if (PRTGIF)
               fprintf (stderr, "%s\n", inbuf);
            else
               Msg (inbuf);
            if (allocated) cfree (line);
            return (FALSE);
         }
      }
      else
      {
         line = inbuf;
         line[len-1] = '\0';
      }

      if (line[0] == '%')
      {
         if (!importingFile && line[1]=='%')
         {
            int	line_len = strlen (line);

            if (savedComments == NULL)
            {
               if ((savedComments = (char *) calloc (line_len+2, sizeof(char)))
                     == NULL)
                  Error ("ReadObj()", "Can not calloc()");
               *savedComments = '\0';
            }
            else
            {
               if ((savedComments = (char *) realloc (savedComments,
                     savedCommentsLen+line_len+2)) == NULL)
                  Error ("ReadObj()", "Can not realloc()");
               savedComments[savedCommentsLen] = '\0';
            }
            strcat (savedComments, line);
            savedCommentsLen += line_len;
            savedComments[savedCommentsLen++] = '\n';
            savedComments[savedCommentsLen] = '\0';
         }
         continue;
      }

      ParseStr (line, '(', obj_name);
      if (strcmp (obj_name, "poly") == 0)
      {
         ReadPolyObj (FP, line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (INVALID, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "box") == 0)
      {
         ReadBoxObj (line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "oval") == 0)
      {
         ReadOvalObj (line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "text") == 0)
      {
         ReadTextObj (FP, line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "polygon") == 0)
      {
         ReadPolygonObj (FP, line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "arc") == 0)
      {
         ReadArcObj (line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "rcbox") == 0)
      {
         ReadRCBoxObj (line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "xbm") == 0)
      {
         ReadXBmObj (FP, line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "xpm") == 0)
      {
         ReadXPmObj (FP, line, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (START_HAVING_ATTRS-1, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "group") == 0)
      {
         ReadGroupObj (FP, OBJ_GROUP, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (INVALID, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "sym") == 0)
      {
         ReadGroupObj (FP, OBJ_SYM, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (INVALID, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "icon") == 0)
      {
         ReadGroupObj (FP, OBJ_ICON, ObjPtr);
         if (*ObjPtr == NULL) return (FALSE);
         ReadObjAttrs (INVALID, FP, ObjPtr);
         AdjObjBBox (*ObjPtr);
         if (allocated) cfree (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "page") == 0)
      {
         *ObjPtr = NULL;
         if (allocated) cfree (line);
         if (importingFile && !importingIconFile && readingPageNum == 1)
         {  /* as if read to the end of file */
            if (!PRTGIF && !pastingFile)
               Msg ("Only page 1 is imported from a multipage file.");
            return (FALSE);
         }
         ReadPageObj (line);
         return (TRUE);
      }
      else if (strcmp (obj_name, "state") == 0)
      {
         read_state_ok = ReadState (line);
         *ObjPtr = NULL;
         if (allocated) cfree (line);
         return ((read_state_ok) ? TRUE : INVALID);
      }
      else if (strcmp (obj_name, "file_attr") == 0)
      {
         if (importingFile && !importingIconFile)
         {
            struct AttrRec	* saved_first_attr, * saved_last_attr;

            saved_first_attr = tgifObj->fattr;
            saved_last_attr = tgifObj->lattr;
            tgifObj->fattr = NULL;
            tgifObj->lattr = NULL;

            ReadObjAttrs (START_HAVING_ATTRS-1, FP, &tgifObj);
            DelAllAttrs (tgifObj->fattr);

            tgifObj->fattr = saved_first_attr;
            tgifObj->lattr = saved_last_attr;
         }
         else
            ReadObjAttrs (START_HAVING_ATTRS-1, FP, &tgifObj);
         if (allocated) cfree (line);
         return (TRUE);
      }
      if (allocated) cfree (line);
   }
   return (FALSE);
}

static
void ConvertToUpperCase (InStr, OutStr)
   register char	* InStr, * OutStr;
{
   for ( ; *InStr != '\0'; InStr++, OutStr++)
      *OutStr = (*InStr>='a' && *InStr<='z') ? *InStr-'a'+'A' : *InStr;
   *OutStr = '\0';
}

void ChangeDomain ()
{
   char 	domain_name[MAXPATHLENGTH+1], env_str[MAXPATHLENGTH+1];
   char 	s[MAXSTRING+1], s1[MAXSTRING+1], * c_ptr;
   char 	cap_tool_name[MAXSTRING+1];
   int		index;
   XEvent	ev;

   index = SelectDomain (domain_name);
   if (index == INVALID) return;

   if (XCheckMaskEvent (mainDisplay, ExposureMask, &ev))
      ExposeEventHandler (&ev, TRUE);

   if (domainInResource)
   {
      sprintf (s, "DomainPath%1d", index);
      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,s)) != NULL)
      {
         char	* c_ptr1;

         while (*c_ptr==' ' || *c_ptr=='\t' || *c_ptr=='\n') c_ptr++;
         if (*c_ptr != '\0' && (c_ptr1=strchr (c_ptr, ':')) != NULL)
            c_ptr = &c_ptr1[1];
         ParseSymPath (c_ptr);
      }
      else if (strcmp (domain_name, "Examples") == 0)
         ParseSymPath (TGIF_PATH);
      else
         ParseSymPath (".");
   }
   else
   {
      ConvertToUpperCase (TOOL_NAME, cap_tool_name);
      sprintf (env_str, "%s_%s", cap_tool_name, domain_name);
      if ((c_ptr = getenv (env_str)) == NULL)
      {
         if (strcmp (domain_name, "Examples") == 0)
            ParseSymPath (TGIF_PATH);
         else
            ParseSymPath (".");
      }
      else
         ParseSymPath (c_ptr);
   }

   UpdateSymInfo ();

   strcpy (curDomainName, domain_name);
   sprintf (s, "Current domain is '%s'.", curDomainName);
   sprintf (s1, "Symbol path set to '%s'.", curDomainPath);
   TwoLineMsg (s, s1);
   RedrawTitleWindow ();
}

void AdjForOldVersion (obj_ptr)
   struct ObjRec	* obj_ptr;
{
   if (fileVersion <= 13)
   {
      switch (gridSystem)
      {
         case ENGLISH_GRID:
            MoveObj (obj_ptr, (int)(HALF_INCH), (int)(HALF_INCH));
            break;
         case METRIC_GRID:
            MoveObj (obj_ptr, (int)(2.5*ONE_CM), (int)(2.5*ONE_CM));
            break;
      }
   }
}

void ImportFile ()
{
   struct ObjRec	* obj_ptr, * saved_top_obj, * saved_bot_obj;
   char 		file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1], * rest;
   FILE			* fp;
   int			short_name, read_status;
   XEvent		ev;
   char 		tmp_filename[MAXPATHLENGTH+1];
   int 			tmp_linenum;

   MakeQuiescent ();

   importingFile = TRUE;
   if (importFromLibrary)
   {
      char	name[MAXSTRING+1], path[MAXSTRING+1];

      if (SelectFromLibrary("Please select an object file to IMPORT...",
            OBJ_FILE_EXT, name, path) == INVALID)
      {
         importingFile = FALSE;
         return;
      }
      sprintf (file_name, "%s/%s", path, name);
   }
   else if (SelectFileNameToImport("Please select an object file to IMPORT...",
         OBJ_FILE_EXT, file_name) == INVALID)
   {
      importingFile = FALSE;
      return;
   }

   if (XCheckMaskEvent (mainDisplay, ExposureMask, &ev))
      ExposeEventHandler (&ev, TRUE);

   if ((short_name = IsPrefix (bootDir, file_name, &rest))) ++rest;
   if ((fp = fopen (file_name, "r")) == NULL)
   {
      if (short_name)
         sprintf (s, "Can not import '%s'.", rest);
      else
         sprintf (s, "Can not import '%s'.", file_name);
      Msg (s);
      importingFile = FALSE;
      return;
   }

   strcpy (tmp_filename, scanFileName);
   tmp_linenum = scanLineNum;
   strcpy (scanFileName, (short_name ? rest : file_name));
   scanLineNum = 0;

   saved_top_obj = topObj;
   saved_bot_obj = botObj;
   curPage->top = curPage->bot = topObj = botObj = NULL;

   if (short_name)
      sprintf (s, "Importing '%s' ...", rest);
   else
      sprintf (s, "Importing '%s' ...", file_name);
   Msg (s);

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   numRedrawBBox = 0;
   readingPageNum = 0;
   while ((read_status = ReadObj (fp, &obj_ptr)) == TRUE)
      if (obj_ptr != NULL)
      {
         AdjForOldVersion (obj_ptr);
         UnlockAnObj (obj_ptr);
         AddObj (NULL, topObj, obj_ptr);
      }

   fclose (fp);
   importingFile = FALSE;

   strcpy (scanFileName, tmp_filename);
   scanLineNum = tmp_linenum;

   if (read_status == INVALID)
   {
      sprintf (s, "File version too large (=%1d).  Import aborted!",
            fileVersion);
      Msg (s);
      SetDefaultCursor (mainWindow);
      ShowCursor ();
      return;
   }

   if (topObj != NULL) SetFileModified (TRUE);
   justDupped = FALSE;

   SelAllObj (FALSE);

   if (botObj != NULL)
      botObj->next = saved_top_obj;
   else
      topObj = saved_top_obj;

   if (saved_top_obj != NULL)
   {
      saved_top_obj->prev = botObj;
      botObj = saved_bot_obj;
   }
   curPage->top = topObj;
   curPage->bot = botObj;

   PrepareToRecord (CMD_NEW, NULL, NULL, 0);
   RecordCmd (CMD_NEW, NULL, topSel, botSel, numObjSelected);
   RedrawDrawWindow (botObj);
   HighLightForward ();

   if (!importFromLibrary) SetCurImportDir (file_name);

   if (short_name)
      sprintf (s, "'%s' imported.", rest);
   else
      sprintf (s, "'%s' imported.", file_name);
   Msg (s);
   SetDefaultCursor (mainWindow);
   ShowCursor ();
}

void ImportXBitmapFile ()
{
   char 		file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1], * rest;
   char 		mag_spec[MAXSTRING+1], msg[MAXSTRING+1];
   unsigned int		tmp_w, tmp_h;
   int			rc, x_hot, y_hot, x, y, w, h, short_name;
   int			orig_w, orig_h;
   float		mag;
   Pixmap		orig_bitmap, bitmap;
   XImage		* image=NULL;
   XEvent		ev;
   struct ObjRec	* obj_ptr;

   MakeQuiescent ();

   importingFile = TRUE;
   if (importFromLibrary)
   {
      char	name[MAXSTRING+1], path[MAXSTRING+1];

      if (SelectFromLibrary("Please select an XBitmap file to IMPORT...",
            XBM_FILE_EXT, name, path) == INVALID)
      {
         importingFile = FALSE;
         return;
      }
      sprintf (s, "%s/%s", path, name);
   }
   else if (SelectFileNameToImport("Please select an XBitmap file to IMPORT...",
         XBM_FILE_EXT, s) == INVALID)
   {
      importingFile = FALSE;
      return;
   }

   if (XCheckMaskEvent (mainDisplay, ExposureMask, &ev))
      ExposeEventHandler (&ev, TRUE);

   strcpy (file_name, s);

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);
   rc = XReadBitmapFile (mainDisplay, mainWindow, file_name, &tmp_w, &tmp_h,
         &orig_bitmap, &x_hot, &y_hot);
   orig_w = tmp_w; orig_h = tmp_h;
   SetDefaultCursor (mainWindow);
   ShowCursor ();

   if ((short_name = IsPrefix (bootDir, file_name, &rest))) ++rest;
   if (rc != BitmapSuccess)
   {
      if (short_name)
         sprintf (s, "Can not import XBitmap file '%s'.", rest);
      else
         sprintf (s, "Can not import XBitmap file '%s'.", file_name);
      Msg (s);
      importingFile = FALSE;
      return;
   }

   x = 0;
   y = 0;
   w = orig_w;
   h = orig_h;
   mag = 1.0;
   if (askForXBmSpec)
   {
      sprintf (msg, "%s: [[MAG=]WxH+X+Y] (original size is %1dx%1d)",
         "Please enter geometry spec", orig_w, orig_h);
      Dialog (msg, "( <CR>: accept, <ESC>: continue )", mag_spec);
      if (*mag_spec != '\0')
         ParseCutSpec (mag_spec, orig_w, orig_h, &mag, &x, &y, &w, &h);

      if (x==0 && y==0 && w==orig_w && h==orig_h && mag==1.0)
         bitmap = orig_bitmap;
      else
      {
         orig_w = w;
         orig_h = h;
         if (!ExtractBitmap (orig_bitmap, NULL, x, y, w, h, &bitmap, &image))
         {
            Msg ("Can not allocate extracted bitmap.");
            importingFile = FALSE;
            XFreePixmap (mainDisplay, orig_bitmap);
            return;
         }
         XFreePixmap (mainDisplay, orig_bitmap);
         w = (int)(((float)w) * mag);
         h = (int)(((float)h) * mag);
      }
   }
   else
      bitmap = orig_bitmap;

   obj_ptr = CreateXBmObj (orig_w, orig_h, w, h, bitmap, image);
   PlaceTopObj (obj_ptr);
   AddObj (NULL, topObj, obj_ptr);

   SelectTopObj ();
   RecordNewObjCmd ();
   SetFileModified (TRUE);
   justDupped = FALSE;

   if (!importFromLibrary) SetCurImportDir (file_name);

   if (short_name)
      sprintf (s, "XBitmap file (%1dx%1d) '%s' imported.", orig_w, orig_h,
            rest);
   else
      sprintf (s, "XBitmap file (%1dx%1d) '%s' imported.", orig_w, orig_h,
            file_name);
   Msg (s);
   importingFile = FALSE;
}

void ImportXPixmapFile ()
{
   char 		file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1], * rest;
   int			rc, ncolors, chars_per_pixel, * pixels, short_name;
   int			first_pixel_is_bg, image_w, image_h, w, h;
   Pixmap		pixmap;
   XImage		* image;
   char			* color_char, * * color_str;
   XEvent		ev;
   struct ObjRec	* obj_ptr;

   MakeQuiescent ();

   importingFile = TRUE;
   if (importFromLibrary)
   {
      char	name[MAXSTRING+1], path[MAXSTRING+1];

      if (SelectFromLibrary("Please select an XPixmap file to IMPORT...",
            XPM_FILE_EXT, name, path) == INVALID)
      {
         importingFile = FALSE;
         return;
      }
      sprintf (s, "%s/%s", path, name);
   }
   else if (SelectFileNameToImport("Please select an XPixmap file to IMPORT...",
         XPM_FILE_EXT, s) == INVALID)
   {
      importingFile = FALSE;
      return;
   }

   if (XCheckMaskEvent (mainDisplay, ExposureMask, &ev))
      ExposeEventHandler (&ev, TRUE);

   strcpy (file_name, s);

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);
   rc = MyReadPixmapFile (file_name, &image_w, &image_h, &w, &h, &pixmap,
         &image, &ncolors, &chars_per_pixel, &first_pixel_is_bg, &color_char,
         &color_str, &pixels);
   SetDefaultCursor (mainWindow);
   ShowCursor ();

   if ((short_name = IsPrefix (bootDir, file_name, &rest))) ++rest;
   if (rc != BitmapSuccess)
   {
      if (short_name)
         sprintf (s, "Can not import XPixmap file '%s'.", rest);
      else
         sprintf (s, "Can not import XPixmap file '%s'.", file_name);
      Msg (s);
      importingFile = FALSE;
      return;
   }

   obj_ptr = CreateXPmObj (image_w, image_h, w, h, pixmap, image, ncolors,
         chars_per_pixel, first_pixel_is_bg, color_char, color_str, pixels);
   PlaceTopObj (obj_ptr);
   AddObj (NULL, topObj, obj_ptr);

   SelectTopObj ();
   RecordNewObjCmd ();
   SetFileModified (TRUE);
   justDupped = FALSE;

   if (!importFromLibrary) SetCurImportDir (file_name);

   if (short_name)
      sprintf (s, "XPixmap file (%1dx%1d) '%s' imported.", image_w, image_h,
            rest);
   else
      sprintf (s, "XPixmap file (%1dx%1d) '%s' imported.", image_w, image_h,
            file_name);
   Msg (s);
   importingFile = FALSE;
}

void ImportEPSFile ()
{
   char 		file_name[MAXPATHLENGTH+1], s[MAXPATHLENGTH+1], * rest;
   char 		* * lines=NULL, write_date[32];
   int			rc, short_name, num_lines, epsf_level, image_w, image_h;
   float		llx, lly, urx, ury;
   Pixmap		bitmap;
   XImage		* image=NULL;
   XEvent		ev;
   struct ObjRec	* obj_ptr;

   MakeQuiescent ();

   importingFile = TRUE;
   if (importFromLibrary)
   {
      char	name[MAXSTRING+1], path[MAXSTRING+1];

      if (SelectFromLibrary("Please select an EPS file to IMPORT...",
            EPSF_FILE_EXT, name, path) == INVALID)
      {
         importingFile = FALSE;
         return;
      }
      sprintf (s, "%s/%s", path, name);
   }
   else if (SelectFileNameToImport("Please select an EPS file to IMPORT...",
         EPSF_FILE_EXT, s) == INVALID)
   {
      importingFile = FALSE;
      return;
   }

   if (XCheckMaskEvent (mainDisplay, ExposureMask, &ev))
      ExposeEventHandler (&ev, TRUE);

   strcpy (file_name, s);

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);
   rc = MyReadEPSFile (file_name, &image_w, &image_h, &bitmap, &image,
         &num_lines, &lines, &epsf_level, &llx, &lly, &urx, &ury, write_date);

   if ((short_name = IsPrefix (bootDir, file_name, &rest))) ++rest;
   if (rc != BitmapSuccess)
   {
      SetDefaultCursor (mainWindow);
      ShowCursor ();
      if (short_name)
         sprintf (s, "Can not import EPS file '%s'.", rest);
      else
         sprintf (s, "Can not import EPS file '%s'.", file_name);
      Msg (s);
      importingFile = FALSE;
      return;
   }

   if (short_name)
      obj_ptr = CreateEPSObj (rest, image_w, image_h, bitmap, image,
            num_lines, lines, epsf_level, llx, lly, urx, ury, write_date);
   else
      obj_ptr = CreateEPSObj (file_name, image_w, image_h, bitmap, image,
            num_lines, lines, epsf_level, llx, lly, urx, ury, write_date);

   if (strcmp(defaultEPSScalingStr,"1") != 0)
      ScaleAnEPSObj (obj_ptr, defaultEPSScaling);

   SetDefaultCursor (mainWindow);
   ShowCursor ();

   PlaceTopObj (obj_ptr);
   AddObj (NULL, topObj, obj_ptr);

   SelectTopObj ();
   RecordNewObjCmd ();
   SetFileModified (TRUE);
   justDupped = FALSE;

   if (!importFromLibrary) SetCurImportDir (file_name);

   if (short_name)
      sprintf (s, "EPS file '%s' imported.", rest);
   else
      sprintf (s, "EPS file '%s' imported.", file_name);
   Msg (s);
   importingFile = FALSE;
}

int LoadFile (FullName, ObjFile)
   char	* FullName;
   int	ObjFile; /* equals TRUE if the file is an OBJ file */
{
   struct ObjRec	* obj_ptr;
   char 		file_name[MAXPATHLENGTH+1], msg[MAXPATHLENGTH+1];
   char 		saved_cur_dir[MAXPATHLENGTH+1], * rest;
   int			read_status, short_name;
   FILE			* fp;
   char 		tmp_filename[MAXPATHLENGTH+1];
   int 			tmp_linenum, interrupted;

   if (ObjFile)
      strcpy (saved_cur_dir, curDir);
   else
      strcpy (saved_cur_dir, curSymDir);
   strcpy (file_name, FullName);

   if ((short_name = IsPrefix (bootDir, file_name, &rest))) ++rest;
   if ((fp = fopen (file_name, "r")) == NULL)
   {
      if (short_name)
         sprintf (msg, "Can not open '%s'.", rest);
      else
         sprintf (msg, "Can not open '%s'.", file_name);
      Msg (msg);
      return (FALSE);
   }

   CleanUpComments ();

   strcpy (tmp_filename, scanFileName);
   tmp_linenum = scanLineNum;
   strcpy (scanFileName, (short_name ? rest : file_name));
   scanLineNum = 0;

   TieLooseEnds ();
   CleanUpDrawingWindow ();
   SetFileModified (FALSE);
   if (short_name)
      sprintf (msg, "Loading '%s' ...", rest);
   else
      sprintf (msg, "Loading '%s' ...", file_name);
   Msg (msg);
   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   XClearWindow (mainDisplay, drawWindow);
   somethingHighLighted = FALSE;

   numRedrawBBox = 0;
   ShowInterrupt ();
   interrupted = FALSE;
   readingPageNum = 0;
   loadedCurPageNum = 0;
   while ((read_status = ReadObj (fp, &obj_ptr)) == TRUE)
   {
      if (obj_ptr != NULL)
      {
         AdjForOldVersion (obj_ptr);
         AddObj (NULL, topObj, obj_ptr);
         if (!interrupted && readingPageNum == loadedCurPageNum &&
               (PointInBBox (obj_ptr->x, obj_ptr->y, drawWinBBox) ||
               BBoxIntersect (obj_ptr->bbox, drawWinBBox)))
         {
            if (!DrawObj (drawWindow, obj_ptr)) interrupted = TRUE;
            if (CheckInterrupt ()) interrupted = TRUE;
         }
      }
   }
   if (interrupted) Msg ("User interrupt.  Drawing aborted.");
   HideInterrupt ();

   strcpy (scanFileName, tmp_filename);
   scanLineNum = tmp_linenum;

   if (read_status == INVALID)
   {
      sprintf (msg, "File version too large (=%1d).  Load aborted!",
            fileVersion);
      Msg (msg);
      DrawPaperBoundary ();
      RedrawGridLines ();
      SetDefaultCursor (mainWindow);
      ShowCursor ();
      return (FALSE);
   }

   fclose (fp);
   if (ObjFile)
   {
      SetCurDir (file_name);
      *curSymDir = '\0';
   }
   else
      SetCurSymDir (file_name);
   curFileDefined = TRUE;

   if (loadedCurPageNum <= 0)
   {
      loadedCurPageNum = curPageNum = lastPageNum = 1;
      InitPage ();
      DrawPaperBoundary ();
      RedrawGridLines ();
      RedrawPageLines ();
      RedrawChoiceWindow ();
   }
   GotoPageNum (loadedCurPageNum);

   if (ObjFile)
   {
      if (strcmp (saved_cur_dir, curDir) != 0 && DirInSymPath ("."))
         UpdateSymInfo ();
   }
   else
   {
      if (strcmp (saved_cur_dir, curSymDir) != 0 && DirInSymPath ("."))
         UpdateSymInfo ();
   }

   if (short_name)
      sprintf (msg, "Current file is '%s'.", rest);
   else
      sprintf (msg, "Current file is '%s'.", file_name);
   Msg (msg);

   RedrawTitleWindow ();
   justDupped = FALSE;
   SetDefaultCursor (mainWindow);
   ShowCursor ();

   CleanUpCmds ();

   UpdateAllSubMenus ();

   return (TRUE);
}

void DumpPatFill (FP, Fill, CellSize, BBox, Blanks)
   FILE		* FP;
   int		Fill, CellSize;
   struct BBRec	BBox;
   char		* Blanks;
{
   int	ltx, lty, rbx, rby;

   ltx = ((BBox.ltx % CellSize) == 0) ? BBox.ltx :
         ((BBox.ltx > 0) ? ((int)(BBox.ltx / CellSize))*CellSize :
         ((int)(BBox.ltx / CellSize)-1)*CellSize);
   lty = ((BBox.lty % CellSize) == 0) ? BBox.lty :
         ((BBox.lty > 0) ? ((int)(BBox.lty / CellSize))*CellSize :
         ((int)(BBox.lty / CellSize)-1)*CellSize);
   rbx = ((BBox.rbx % CellSize) == 0) ? BBox.rbx :
         ((BBox.rbx > 0) ? ((int)(BBox.rbx / CellSize)+1)*CellSize :
         ((int)(BBox.rbx / CellSize))*CellSize);
   rby = ((BBox.rby % CellSize) == 0) ? BBox.rby :
         ((BBox.rby > 0) ? ((int)(BBox.rby / CellSize)+1)*CellSize :
         ((int)(BBox.rby / CellSize))*CellSize);

   if (fprintf (FP, "%spat%1d %1d %1d %1d %1d %1d tgifpatfill\n",
         Blanks, Fill, CellSize, ltx, lty, rbx-ltx, rby-lty) == EOF)
      writeFileFailed = TRUE;
}

void DumpSymOutline (FP, ObjPtr)
   FILE				* FP;
   register struct ObjRec	* ObjPtr;
{
   int  ltx, lty, rbx, rby;

   ltx = ObjPtr->obbox.ltx - QUARTER_INCH + 1;
   lty = ObjPtr->obbox.lty - QUARTER_INCH + 1;
   rbx = ObjPtr->obbox.rbx + QUARTER_INCH - 1;
   rby = ObjPtr->obbox.rby + QUARTER_INCH - 1;

   fprintf (FP, "gsave\n");
   fprintf (FP, "   0 setgray\n");
   fprintf (FP, "   [4 4] 0 setdash\n");
   fprintf (FP, "   newpath\n   %1d %1d moveto ", ltx, lty);
   fprintf (FP, "%1d %1d lineto ", rbx, lty);
   fprintf (FP, "%1d %1d lineto ", rbx, rby);
   fprintf (FP, "%1d %1d lineto\n", ltx, rby);
   fprintf (FP, "   closepath stroke\n");
   fprintf (FP, "grestore\n");
}

static int	printingFirstPageNum=1;
static int	printingPageNum=1, printingPageRow=1, printingPageCol=1;
static int	printingLastPageNum=0;
static int	dumpOnePageInTileMode=FALSE;
static int	dumpOnePageInStackMode=FALSE;

static FILE	* dumpFP=NULL;
static char	tmpFile[MAXSTRING+1];
static int	llxTotal=0, llyTotal=0, urxTotal=0, uryTotal=0;
static int	totalBBoxValid=FALSE;

static int	msgAboutTiledPageScalingSeen=FALSE;

static
void DumpAttrs (FP, AttrPtr)
   FILE				* FP;
   register struct AttrRec	* AttrPtr;
{
   for ( ; AttrPtr != NULL; AttrPtr = AttrPtr->prev)
      if (AttrPtr->shown)
      {
         if (!AttrPtr->nameshown && *AttrPtr->name == '!' &&
               strcmp(AttrPtr->name, "!PAGE_NUM=") == 0)
         {
            if (pageLayoutMode == PAGE_STACK)
            {
               struct StrRec	* s_ptr=AttrPtr->obj->detail.t->first;
               char		* c_ptr;

               if (s_ptr != NULL && (c_ptr=strstr(s_ptr->s,
                     "!(STACKED_PAGE_NUM)")) != NULL)
               {
                  char	saved_str[MAXSTRING+1], * c_ptr1;
                  int	len;

                  strcpy(saved_str, s_ptr->s);
                  sprintf(c_ptr, "%1d", printingPageNum);
                  len = strlen(s_ptr->s);
                  c_ptr1 = &c_ptr[19];
                  c_ptr = &s_ptr->s[len];
                  while (*c_ptr1 != '\0') *c_ptr++ = *c_ptr1++;
                  *c_ptr = '\0';
                  DumpTextObj (FP, AttrPtr->obj);
                  strcpy(s_ptr->s, saved_str);
               }
               else
                  DumpTextObj (FP, AttrPtr->obj);
            }
            else
            {
               struct StrRec	* s_ptr=AttrPtr->obj->detail.t->first;
               char		* c_ptr;

               if (s_ptr != NULL)
               {
                  char	saved_str[MAXSTRING+1], * c_ptr1, * c_ptr2;
                  int	len;

                  strcpy(saved_str, s_ptr->s);
                  for (c_ptr=s_ptr->s; *c_ptr != '\0'; c_ptr++)
                  {
                     if (*c_ptr == '!' &&
                           (strncmp(c_ptr, "!(TILED_PAGE_ROW)", 17) == 0 ||
                           strncmp(c_ptr, "!(TILED_PAGE_COL)", 17) == 0))
                     {
                        if (strncmp(c_ptr, "!(TILED_PAGE_ROW)", 17) == 0)
                           sprintf(c_ptr, "%1d", printingPageRow);
                        else
                           sprintf(c_ptr, "%1d", printingPageCol);
                        len = strlen(c_ptr);
                        c_ptr1 = &c_ptr[17];
                        c_ptr = c_ptr2 = &c_ptr[len];
                        while (*c_ptr1 != '\0') *c_ptr2++ = *c_ptr1++;
                        *c_ptr2 = '\0';
                        c_ptr--;
                     }
                  }
                  DumpTextObj (FP, AttrPtr->obj);
                  strcpy(s_ptr->s, saved_str);
               }
               else
                  DumpTextObj (FP, AttrPtr->obj);
            }
         }
         else
            DumpTextObj (FP, AttrPtr->obj);
      }
}

static
void DumpAnObj (FP, ObjPtr)
   FILE				* FP;
   register struct ObjRec	* ObjPtr;
{
   register struct ObjRec	* obj_ptr;

   switch (ObjPtr->type)
   {
      case OBJ_POLY:
         DumpPolyObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_BOX:
         DumpBoxObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_OVAL:
         DumpOvalObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_TEXT: DumpTextObj (FP, ObjPtr); break;
      case OBJ_POLYGON:
         DumpPolygonObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_ARC:
         DumpArcObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_RCBOX:
         DumpRCBoxObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_XBM:
         DumpXBmObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_XPM:
         DumpXPmObj (FP, ObjPtr);
         DumpAttrs (FP, ObjPtr->lattr);
         break;
      case OBJ_SYM:
      case OBJ_ICON:
      case OBJ_GROUP:
         obj_ptr = ObjPtr->detail.r->last;
         for ( ; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
            DumpAnObj (FP, obj_ptr);
         DumpAttrs (FP, ObjPtr->lattr);
         if (ObjPtr->type == OBJ_SYM) DumpSymOutline (FP, ObjPtr);
         break;
   }
}

static int	llxPage=0, llyPage=0, urxPage=0, uryPage=0;

#define ALL_BBOX 0
#define PAGE_BBOX 1

int DumpBBox (fp, page_only, page_bbox)
   FILE		* fp;
   int		page_only;
   struct BBRec	* page_bbox;
{
   register struct ObjRec	* obj_ptr;
   int				ltx=0, lty=0, rbx=0, rby=0, rc=TRUE;
   double			llx1=0, lly1=0, urx1=0, ury1=0;

   if ((obj_ptr = topObj) == NULL)
   {
      char msg[MAXSTRING];

      ltx = lty = rbx = rby = 0;
      if (!(pageLayoutMode == PAGE_STACK &&
            printingFirstPageNum != printingLastPageNum))
      {
         strcpy (msg, "Warning:  The PostScript bounding box is empty!");
         if (PRTGIF)
            fprintf (stderr, "%s\n", msg);
         else
            Msg (msg);
      }
      rc = FALSE;
   }
   else if (pageLayoutMode == PAGE_STACK || page_bbox == NULL)
   {
      ltx = obj_ptr->bbox.ltx; lty = obj_ptr->bbox.lty;
      rbx = obj_ptr->bbox.rbx; rby = obj_ptr->bbox.rby;

      for (obj_ptr = topObj->next; obj_ptr != NULL; obj_ptr = obj_ptr->next)
      {
         if (obj_ptr->bbox.ltx < ltx) ltx = obj_ptr->bbox.ltx;
         if (obj_ptr->bbox.lty < lty) lty = obj_ptr->bbox.lty;
         if (obj_ptr->bbox.rbx > rbx) rbx = obj_ptr->bbox.rbx;
         if (obj_ptr->bbox.rby > rby) rby = obj_ptr->bbox.rby;
      }
   }
   else
   {
      int found=FALSE;

      for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
         if (Inside (obj_ptr->bbox, *page_bbox) ||
               BBoxIntersect (obj_ptr->bbox, *page_bbox))
         {
            if (found)
            {
               if (obj_ptr->bbox.ltx < ltx) ltx = obj_ptr->bbox.ltx;
               if (obj_ptr->bbox.lty < lty) lty = obj_ptr->bbox.lty;
               if (obj_ptr->bbox.rbx > rbx) rbx = obj_ptr->bbox.rbx;
               if (obj_ptr->bbox.rby > rby) rby = obj_ptr->bbox.rby;
            }
            else
            {
               found = TRUE;
               ltx = obj_ptr->bbox.ltx; lty = obj_ptr->bbox.lty;
               rbx = obj_ptr->bbox.rbx; rby = obj_ptr->bbox.rby;
            }
         }
      if (!found)
      {
         rc = FALSE;
         ltx = lty = rbx = rby = 0;
      }
      else if (page_bbox != NULL)
      {
         struct BBRec	bbox, bbox1;

         bbox.ltx = ltx; bbox.lty = lty; bbox.rbx = rbx; bbox.rby = rby;
         if (IntersectRect (bbox, *page_bbox, &bbox1))
         {
            ltx = bbox1.ltx-page_bbox->ltx; lty = bbox1.lty-page_bbox->lty;
            rbx = bbox1.rbx-page_bbox->ltx; rby = bbox1.rby-page_bbox->lty;
         }
      }
   }
   switch (pageStyle)
   {
      case PORTRAIT:
         llx1 = (double) (1.0*ltx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
               psXOff[pageStyle]*psDotsPerInch);
         lly1 = (double) (-1.0*rby*psDotsPerInch/PIX_PER_INCH*printMag/100 +
               psYOff[pageStyle]*psDotsPerInch);
         urx1 = (double) (1.0*rbx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
               psXOff[pageStyle]*psDotsPerInch);
         ury1 = (double) (-1.0*lty*psDotsPerInch/PIX_PER_INCH*printMag/100 +
               psYOff[pageStyle]*psDotsPerInch);
         break;
      case LANDSCAPE:
         llx1 = (double) (1.0*lty*psDotsPerInch/PIX_PER_INCH*printMag/100 -
               psYOff[pageStyle]*psDotsPerInch);
         lly1 = (double) (1.0*ltx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
               psXOff[pageStyle]*psDotsPerInch);
         urx1 = (double) (1.0*rby*psDotsPerInch/PIX_PER_INCH*printMag/100 -
               psYOff[pageStyle]*psDotsPerInch);
         ury1 = (double) (1.0*rbx*psDotsPerInch/PIX_PER_INCH*printMag/100 +
               psXOff[pageStyle]*psDotsPerInch);
         break;
      default:
         fprintf (stderr, "Unrecognizable page style '%1d'\n", pageStyle);
         break;
   }

   if (rc)
   {
      llxPage = (llx1 >= 0.0) ? ((int)llx1)-1 : (-((int)(-llx1)))-1;
      llyPage = (lly1 >= 0.0) ? ((int)lly1)-1 : (-((int)(-lly1)))-1;
      urxPage = (urx1 >= 0.0) ? ((int)urx1)+1 : (-((int)(0.999-urx1)))+1;
      uryPage = (ury1 >= 0.0) ? ((int)ury1)+1 : (-((int)(0.999-ury1)))+1;
   }
   else
      llxPage = llyPage = urxPage = uryPage = 0;

   fprintf (fp,"%%%%%sBoundingBox: %1d %1d %1d %1d\n",
         (page_only ? "Page" : ""), llxPage, llyPage, urxPage, uryPage);
   return (rc);
}

void ModifyOutputFileName (FileName)
   char	* FileName;
{
   register int	i;
   int		len;
   char		s[MAXPATHLENGTH+1];

   if (*outputDir == '\0') return;

   strcpy (s, FileName);
   len = strlen (s);
   for (i = len-1; i >= 0 && s[i] != '/'; i--) ;
   if (i >= 0)
      sprintf (FileName, "%s/%s", outputDir, &s[i+1]);
   else
      sprintf (FileName, "%s/%s", outputDir, s);
}

static char	psBopHook[MAXSTRING], psEopHook[MAXSTRING];
static int	psBopHookStatus=INVALID, psEopHookStatus=INVALID;
static int	minimalEPS=INVALID;

static
void GenDump (FileName)
   char	* FileName;
{
   register struct ObjRec	* obj_ptr;
   char				cmd[MAXSTRING+1], tmp_str[MAXSTRING+1];
   char				ps_file[MAXSTRING+1], * c_ptr;
   char				msg[MAXSTRING+1], * rest, loc_time[MAXSTRING+1];
   int				i, len, short_name = FALSE;
   time_t			tloc;
   struct DocFontRec		* next_df;

   if (PRTGIF)
   {
      minimalEPS = TRUE;
   }
   else if (minimalEPS == INVALID)
   {
      minimalEPS = TRUE;
      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"MinimalEPS")) != NULL &&
            (strcmp (c_ptr, "false") == 0 || strcmp (c_ptr, "False") == 0))
         minimalEPS = FALSE;
   }
   if (!PRTGIF && psBopHookStatus == INVALID)
   {
      *psBopHook = '\0';
      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PSBopHook")) != NULL)
      {
         strcpy (psBopHook, c_ptr);
         psBopHookStatus = TRUE;
      }
      else
         psBopHookStatus = FALSE;
   }
   if (!PRTGIF && psEopHookStatus == INVALID)
   {
      *psEopHook = '\0';
      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"PSEopHook")) != NULL)
      {
         strcpy (psEopHook, c_ptr);
         psEopHookStatus = TRUE;
      }
      else
         psEopHookStatus = FALSE;
   }

   if (botObj==NULL && (pageLayoutMode==PAGE_STACK && dumpOnePageInStackMode ||
         pageLayoutMode==PAGE_TILE))
   {
      sprintf (msg, "No objects to print.");
      if (PRTGIF)
         fprintf (stderr, "%s\n", msg);
      else
         Msg (msg);
      return;
   }
   if (!PRTGIF && printingPageNum==printingFirstPageNum)
      Msg ("Generating print file ...");

   if (whereToPrint == XBM_FILE)
   {
      if (!curFileDefined)
      {
         sprintf (msg, "No current file.  Can not generate X11 %s output!",
               (colorDump ? "pixmap" : "bitmap"));
         if (PRTGIF)
            fprintf (stderr, "%s\n", msg);
         else
            Dialog (msg, NULL, NULL);
      }
      else if (topObj == NULL)
      {
         sprintf (msg, "No objects to print.");
         if (PRTGIF)
            fprintf (stderr, "%s\n", msg);
         else
            Dialog (msg, NULL, NULL);
      }
      else
      {
         SetWatchCursor (drawWindow);
         SetWatchCursor (mainWindow);
         DumpXBitmapFile ();
         SetDefaultCursor (mainWindow);
         ShowCursor ();
      }
      return;
   }

   if (printingPageNum == printingFirstPageNum)
   {
      sprintf (tmpFile, "%sTgifXXXXXX", TMP_DIR);
      mktemp (tmpFile);
      unlink (tmpFile);

      if ((dumpFP = fopen (tmpFile, "w")) == NULL)
      {
         sprintf (tmp_str, "Can not create '%s', print aborted.", tmpFile);
         if (PRTGIF)
            fprintf (stderr, "%s\n", tmp_str);
         else
            Msg (tmp_str);
         return;
      }

      if (PRTGIF) fprintf (stderr, "Writing to '%s' ...\n", tmpFile);

      writeFileFailed = FALSE;
      ResetGrayDetection ();

      if (usePsAdobeString)
      {
         switch (whereToPrint)
         {
            case PRINTER:
            case PS_FILE:
               if (*adobeString == '\0')
               {
                  if (fprintf (dumpFP, "%%!PS-Adobe-2.0\n") == EOF)
                     writeFileFailed = TRUE;
               }
               else if (fprintf (dumpFP,"%%!PS-Adobe-%s\n",adobeString) == EOF)
                  writeFileFailed = TRUE;
               break;
            case LATEX_FIG:
               if (*adobeString == '\0')
               {
                  if (fprintf (dumpFP, "%%!PS-Adobe-2.0 EPSF-1.2\n") == EOF)
                     writeFileFailed = TRUE;
               }
               else if (*epsfString == '\0')
               {
                  if (fprintf (dumpFP, "%%!PS-Adobe-%s\n", adobeString) == EOF)
                     writeFileFailed = TRUE;
               }
               else
               {
                  if (fprintf (dumpFP, "%%!PS-Adobe-%s EPSF-%s\n",
                        adobeString, epsfString) == EOF)
                     writeFileFailed = TRUE;
               }
               break;
         }
      }
      else if (fprintf (dumpFP, "%%!\n") == EOF)
            writeFileFailed = TRUE;
      if (pageLayoutMode == PAGE_STACK &&
            printingFirstPageNum == printingLastPageNum)
         DumpBBox (dumpFP, ALL_BBOX, NULL);
      else if (fprintf (dumpFP,"%%%%BoundingBox: (atend)\n") == EOF)
            writeFileFailed = TRUE;
      if (PRTGIF)
      {
         if (fprintf (dumpFP, "%%%%Title: %s\n", FileName) == EOF)
            writeFileFailed = TRUE;
      }
      else if (curFileDefined)
      {
         strcpy (tmp_str, curFileName);
         len = strlen (tmp_str);
         for (i = len-1; tmp_str[i] != '.'; i--) ;
         tmp_str[i] = '\0';
         if (fprintf (dumpFP, "%%%%Title: %s\n", tmp_str) == EOF)
            writeFileFailed = TRUE;
      }
      else if (fprintf (dumpFP, "%%%%Title: [Unnamed]\n") == EOF)
         writeFileFailed = TRUE;
      time (&tloc);
      strcpy(loc_time, ctime (&tloc));
      loc_time[24] = '\0';
      if (fprintf (dumpFP, "%%%%CreationDate: %s\n", loc_time) == EOF)
         writeFileFailed = TRUE;
      if (TGIF_PATCHLEVEL == 0)
      {
         if (fprintf (dumpFP, "%%%%Creator: %s-%s by %s\n",
               TOOL_NAME, version_string,
               "William Chia-Wei Cheng (william@cs.UCLA.edu)") == EOF)
            writeFileFailed = TRUE;
      }
      else
      {
         if (fprintf (dumpFP, "%%%%Creator: %s-%s-p%1d by %s\n",
               TOOL_NAME, version_string, TGIF_PATCHLEVEL,
               "William Chia-Wei Cheng (william@cs.UCLA.edu)") == EOF)
            writeFileFailed = TRUE;
      }
      if (whereToPrint != LATEX_FIG || !minimalEPS)
      {
         if (fprintf (dumpFP, "%%%%Pages: %1d\n", 
               pageLayoutMode == PAGE_STACK ? lastPageNum :
               paperCol * paperRow) == EOF)
            writeFileFailed = TRUE;
         if (fprintf (dumpFP, "%%%%DocumentFonts: (atend)\n") == EOF)
            writeFileFailed = TRUE;
         if (fprintf (dumpFP, "%%%%EndComments\n") == EOF)
            writeFileFailed = TRUE;
         if (fprintf (dumpFP, "%%%%BeginProlog\n") == EOF)
            writeFileFailed = TRUE;
      }

      DumpPSMacro (dumpFP);

      PrepareEightBitFontInfo ();
      DumpEightBitFontInfo (dumpFP);

      if (fprintf (dumpFP, "end\n\n") == EOF) writeFileFailed = TRUE;
      if (whereToPrint != LATEX_FIG || !minimalEPS)
      {
         if (fprintf (dumpFP, "%%%%EndProlog\n") == EOF) writeFileFailed = TRUE;
      }
   }
   if (pageLayoutMode == PAGE_STACK || (pageLayoutMode == PAGE_TILE &&
         (paperCol == 1 && paperRow == 1 || whereToPrint == LATEX_FIG &&
         !dumpOnePageInTileMode)))
   {
      printingPageRow = printingPageCol = 1;
      if (whereToPrint != LATEX_FIG || !minimalEPS)
      {
         fprintf (dumpFP, "%%%%Page: %1d %1d\n\n",
               printingPageNum-printingFirstPageNum+1,
               printingPageNum-printingFirstPageNum+1);
      }
      DumpBBox (dumpFP, PAGE_BBOX, NULL);
      if (totalBBoxValid)
      {
         if (topObj != NULL)
         {
            if (llxPage < llxTotal) llxTotal = llxPage;
            if (llyPage < llyTotal) llyTotal = llyPage;
            if (urxPage > urxTotal) urxTotal = urxPage;
            if (uryPage > uryTotal) uryTotal = uryPage;
         }
      }
      else if (topObj != NULL)
      {
         totalBBoxValid = TRUE;
         llxTotal = llxPage; llyTotal = llyPage;
         urxTotal = urxPage; uryTotal = uryPage;
      }
      if (psBopHookStatus == TRUE && (whereToPrint == PRINTER ||
            whereToPrint == PS_FILE))
         fprintf (dumpFP, "userdict /%s known { %s } if\n\n",
               psBopHook, psBopHook);
      fprintf (dumpFP, "tgifdict begin\n");
      fprintf (dumpFP, "/tgifsavedpage save def\n\n");
      fprintf (dumpFP, "1 setmiterlimit\n");
      fprintf (dumpFP, "1 setlinewidth\n\n");

      if (pageStyle == LANDSCAPE) fprintf (dumpFP, "90 rotate\n");

      fprintf (dumpFP, "%1d %s mul %1d %s mul translate\n", psDotsPerInch,
            psXOffStr[pageStyle], psDotsPerInch, psYOffStr[pageStyle]);

      fprintf (dumpFP, "%1d %1d div %1d mul 100 div dup neg scale\n\n",
            psDotsPerInch, PIX_PER_INCH, printMag);

      fprintf (dumpFP, "gsave\n\n");

      for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
         DumpAnObj (dumpFP, obj_ptr);

      fprintf (dumpFP, "grestore\n");
      fprintf (dumpFP, "tgifsavedpage restore\n");
      fprintf (dumpFP, "end\n");

      switch (whereToPrint)
      {
         case PRINTER:
         case PS_FILE:
            if (psEopHookStatus == TRUE)
               fprintf (dumpFP,"userdict /%s known { %s } if\n",
                     psEopHook, psEopHook);
            fprintf (dumpFP, "showpage\n\n");
            break;
         case LATEX_FIG: break;
      }
   }
   else
   {
      int	row, col;
      float	one=(float)atof("1.0"), f_rbx, f_rby, x_dist, y_dist;

      if (tiledPageScaling == one && !msgAboutTiledPageScalingSeen)
      {
         char	msg1[MAXSTRING+1];

         msgAboutTiledPageScalingSeen = TRUE;
         sprintf (msg, "%s %s*TiledPageScaling is set to 1.",
               "Output may get truncated because", TOOL_NAME);
         sprintf (msg1, "    Reducing the above setting is recommended.");
         if (PRTGIF)
            fprintf (stderr, "%s\n%s\n", msg, msg1);
         else
            TwoLineMsg (msg, msg1);
      }
      f_rbx = (float)(((float)PIX_PER_INCH) * 100.0 / ((float)printMag) *
            psPageWidthInInch[pageStyle]);
      f_rby = (float)(((float)PIX_PER_INCH) * 100.0 / ((float)printMag) *
            psPageHeightInInch[pageStyle]);
      x_dist = (float)(((float)PIX_PER_INCH) * psPageWidthInInch[pageStyle] *
            (1.0-tiledPageScaling) * 50.0 / ((float)printMag));
      y_dist = (float)(((float)PIX_PER_INCH) * psPageHeightInInch[pageStyle] *
            (1.0-tiledPageScaling) * 50.0 / ((float)printMag));

      for (row = 0; row < paperRow; row++)
      {
         printingPageRow = row+1;
         for (col = 0; col < paperCol; col++)
         {
            struct BBRec	page_bbox;
            float		real_ltx, real_lty;

            if (dumpOnePageInTileMode && row*paperCol+col+1 != printingPageNum)
               continue;

            printingPageCol = col+1;
            page_bbox.ltx = col * onePageWidth;
            page_bbox.lty = row * onePageHeight;
            page_bbox.rbx = page_bbox.ltx + onePageWidth;
            page_bbox.rby = page_bbox.lty + onePageHeight;

            if (whereToPrint != LATEX_FIG || !minimalEPS)
            {
               if (dumpOnePageInTileMode)
                  fprintf (dumpFP, "%%%%Page: 1 1\n\n");
               else
                  fprintf (dumpFP, "%%%%Page: %1d %1d\n\n", col+1+row*paperCol,
                        col+1+row*paperCol);
            }

            if (DumpBBox (dumpFP, PAGE_BBOX, &page_bbox))
            {
               if (totalBBoxValid)
               {
                  if (topObj != NULL)
                  {
                     if (llxPage < llxTotal) llxTotal = llxPage;
                     if (llyPage < llyTotal) llyTotal = llyPage;
                     if (urxPage > urxTotal) urxTotal = urxPage;
                     if (uryPage > uryTotal) uryTotal = uryPage;
                  }
               }
               else if (topObj != NULL)
               {
                  totalBBoxValid = TRUE;
                  llxTotal = llxPage; llyTotal = llyPage;
                  urxTotal = urxPage; uryTotal = uryPage;
               }
            }
            if (psBopHookStatus == TRUE && (whereToPrint == PRINTER ||
                  whereToPrint == PS_FILE))
               fprintf (dumpFP, "userdict /%s known { %s } if\n\n",
                     psBopHook, psBopHook);
            fprintf (dumpFP, "tgifdict begin\n");
            fprintf (dumpFP, "/tgifsavedpage save def\n\n");
            fprintf (dumpFP, "1 setmiterlimit\n");
            fprintf (dumpFP, "1 setlinewidth\n\n");

            if (pageStyle == LANDSCAPE) fprintf (dumpFP, "90 rotate\n");

            fprintf (dumpFP, "%1d %s mul %1d %s mul translate\n", psDotsPerInch,
                  psXOffStr[pageStyle], psDotsPerInch, psYOffStr[pageStyle]);

            fprintf (dumpFP, "%1d %1d div %1d mul 100 div dup neg scale\n\n",
                  psDotsPerInch, PIX_PER_INCH, printMag);

            fprintf (dumpFP, "gsave\n\n");

            fprintf (dumpFP, "%% SETUP CLIP FOR PAGE\n\n");
            fprintf (dumpFP, "%.3f %.3f translate\n", x_dist, y_dist);
            fprintf (dumpFP, "%.3f %.3f scale\n", tiledPageScaling,
                  tiledPageScaling);

            fprintf (dumpFP, "-%1d -%1d translate\n", page_bbox.ltx,
                  page_bbox.lty);
            real_ltx = (float)(((float)col) *
                  ((float)psPageWidthInInch[pageStyle]) *
                  ((float)PIX_PER_INCH) * 100.0 / ((float)printMag));
            real_lty = (float)(((float)row) *
                  ((float)psPageHeightInInch[pageStyle]) *
                  ((float)PIX_PER_INCH) * 100.0 / ((float)printMag));
            fprintf (dumpFP, "newpath\n");
            fprintf (dumpFP, "   %.3f %.3f moveto\n", real_ltx, real_lty);
            fprintf (dumpFP, "   %.3f 0 rlineto\n", f_rbx);
            fprintf (dumpFP, "   0 %.3f rlineto\n", f_rby);
            fprintf (dumpFP, "   -%.3f 0 rlineto\n", f_rbx);
            fprintf (dumpFP, "closepath clip newpath\n\n");

            for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
               if (Inside (obj_ptr->bbox, page_bbox) ||
                     BBoxIntersect (obj_ptr->bbox, page_bbox))
                  DumpAnObj (dumpFP, obj_ptr);

            fprintf (dumpFP, "grestore\n");
            fprintf (dumpFP, "tgifsavedpage restore\n");
            fprintf (dumpFP, "end\n");

            switch (whereToPrint)
            {
               case PRINTER:
               case PS_FILE:
                  if (psEopHookStatus == TRUE)
                     fprintf (dumpFP, "userdict /%s known { %s } if\n",
                           psEopHook, psEopHook);
                  fprintf (dumpFP, "showpage\n\n");
                  break;
            }
         }
      }
   }
   if (printingPageNum != printingLastPageNum) return;

   if (whereToPrint != LATEX_FIG || !minimalEPS)
   {
      if (fprintf (dumpFP, "%%%%Trailer\n") == EOF)
         writeFileFailed = TRUE;
   }
   if (pageLayoutMode!=PAGE_STACK || printingFirstPageNum!=printingLastPageNum)
      if (fprintf (dumpFP,"%%%%BoundingBox: %1d %1d %1d %1d\n",
            llxTotal, llyTotal, urxTotal, uryTotal) == EOF)
         writeFileFailed = TRUE;
   if (fprintf (dumpFP, "%%MatchingCreationDate: %s\n", loc_time) == EOF)
      writeFileFailed = TRUE;
   if (whereToPrint != LATEX_FIG || !minimalEPS)
   {
      if (fprintf (dumpFP, "%%%%DocumentFonts: ") == EOF)
         writeFileFailed = TRUE;
      for ( ; firstDocFont != NULL; firstDocFont = next_df)
      {
         next_df = firstDocFont->next;
         if (fprintf (dumpFP, "%s", firstDocFont->name) == EOF)
            writeFileFailed = TRUE;
         if (next_df != NULL)
            if (fprintf (dumpFP, "\n%%%%+ ") == EOF)
               writeFileFailed = TRUE;
         cfree (firstDocFont);
      }
      if (fprintf (dumpFP, "\n") == EOF) writeFileFailed = TRUE;
      if (fprintf (dumpFP, "%%%%EOF\n") == EOF) writeFileFailed = TRUE;
   }
   fclose (dumpFP);

   EndGrayDetection ();

   if (!totalBBoxValid)
   {
      sprintf (msg, "No objects to print.");
      if (PRTGIF)
         fprintf (stderr, "%s\n", msg);
      else
         Msg (msg);
      return;
   }
   if (writeFileFailed)
   {
      writeFileFailed = FALSE;
      sprintf (msg, "Fail to write to '%s'.  File system may be full.",
            tmpFile);
      if (PRTGIF)
         fprintf (stderr, "%s\n", msg);
      else
         Dialog (msg, NULL, NULL);
#ifdef KEEP_WHEN_PRINT
      if (whereToPrint != PRINTER) unlink (tmpFile);
#else
      unlink (tmpFile);
#endif
      return;
   }

   switch (whereToPrint)
   {
      case PRINTER:
#ifdef VMS
#define PRINT_TRAILER ""
#else
#define PRINT_TRAILER " 2>&1"
#endif
         if (PRTGIF)
         {
            if (lastFile)
               sprintf (cmd,"%s %s%s",printCommand,tmpFile,PRINT_TRAILER);
            else
               sprintf (cmd,"%s -h %s%s",printCommand,tmpFile,PRINT_TRAILER);
            fprintf (stderr, "%s\n", cmd);
         }
         else
         {
            sprintf (cmd,"%s %s%s",printCommand,tmpFile,PRINT_TRAILER);
            sprintf (msg, "Printing with '%s' command.", printCommand);
            Msg (msg);
         }
         if (!ExecuteCmd (cmd))
         {
            if (PRTGIF)
               fprintf (stderr, "Can not execute '%s', print aborted.\n", cmd);
            else
            {
               sprintf (msg, "Can not execute '%s', print aborted.", cmd);
               Msg (msg);
            }
            unlink (tmpFile);
            return;
         }
         if (PRTGIF)
            fprintf (stderr, "'%s' printed.\n\n", tmpFile);
         else
            Msg ("Print completed.");
         break;
      case LATEX_FIG:
         if (PRTGIF)
         {
            sprintf (ps_file, "%s.%s", FileName, EPSF_FILE_EXT);
            ModifyOutputFileName (ps_file);
         }
         else
         {
            if (!curFileDefined)
            {
               Dialog ("No current file.  Can not generate LaTeX output!",
                     "( <CR> or <ESC> to continue )", cmd);
               unlink (tmpFile);
               return;
            }
            sprintf (ps_file, "%s/%s", curDir, curFileName);
            len = strlen (ps_file);
            for (i = len-1; ps_file[i] != '.'; i--) ;
            sprintf (&ps_file[i], ".%s", EPSF_FILE_EXT);
            ModifyOutputFileName (ps_file);
            if ((short_name = IsPrefix (bootDir, ps_file, &rest))) ++rest;
            if (short_name && *outputDir=='\0')
               sprintf (cmd, "Printing into '%s' ...", rest);
            else
               sprintf (cmd, "Printing into '%s' ...", ps_file);
            Msg (cmd);
         }
         if (!CopyAFile (tmpFile, ps_file))
         {
            if (PRTGIF)
               fprintf (stderr, "LaTeX output not generated.\n");
            else
               Msg ("LaTeX output not generated.");
            unlink (tmpFile);
            return;
         }
         if (PSFILE_MOD != 0 && chmod (ps_file, PSFILE_MOD))
         {
            if (PRTGIF)
               fprintf (stderr, "Can not chmod '%s' to 0%1o.\n", ps_file,
                     PSFILE_MOD);
            else
            {
               if (short_name && *outputDir=='\0')
                  sprintf(msg,"Can not chmod '%s' to 0%1o.",rest,PSFILE_MOD);
               else
                  sprintf(msg,"Can not chmod '%s' to 0%1o.",ps_file,PSFILE_MOD);
               Msg (msg);
            }
         }

         if (PRTGIF)
            fprintf (stderr, "LaTeX figure printed into '%s'.\n\n", ps_file);
         else
         {
            if (short_name && *outputDir=='\0')
               sprintf (msg, "LaTeX figure printed into '%s'.", rest);
            else
               sprintf (msg, "LaTeX figure printed into '%s'.", ps_file);
            Msg (msg);
         }
         break;
      case PS_FILE:
         if (PRTGIF)
         {
            sprintf (ps_file, "%s.%s", FileName, PS_FILE_EXT);
            ModifyOutputFileName (ps_file);
         }
         else
         {
            if (!curFileDefined)
            {
               Dialog ("No current file.  Can not generate PostScript output!",
                     "( <CR> or <ESC> to continue )", cmd);
               unlink (tmpFile);
               return;
            }
            sprintf (ps_file, "%s/%s", curDir, curFileName);
            len = strlen (ps_file);
            for (i = len-1; ps_file[i] != '.'; i--) ;
            sprintf (&ps_file[i], ".%s", PS_FILE_EXT);
            ModifyOutputFileName (ps_file);
            if ((short_name = IsPrefix (bootDir, ps_file, &rest))) ++rest;
            if (short_name && *outputDir=='\0')
               sprintf (cmd, "Printing into '%s' ...", rest);
            else
               sprintf (cmd, "Printing into '%s' ...", ps_file);
            Msg (cmd);
         }
         if (!CopyAFile (tmpFile, ps_file))
         {
            if (PRTGIF)
               fprintf (stderr, "PostScript output not generated.\n");
            else
               Msg ("PostScript output not generated.");
            unlink (tmpFile);
            return;
         }
         if (PSFILE_MOD != 0 && chmod (ps_file, PSFILE_MOD))
         {
            if (PRTGIF)
               fprintf (stderr, "Can not chmod '%s' to 0%1o.\n", ps_file,
                     PSFILE_MOD);
            else
            {
               if (short_name && *outputDir=='\0')
                  sprintf(msg,"Can not chmod '%s' to 0%1o.",rest,PSFILE_MOD);
               else
                  sprintf(msg,"Can not chmod '%s' to 0%1o.",ps_file,PSFILE_MOD);
               Msg (msg);
            }
         }
         if (PRTGIF)
            fprintf (stderr, "PostScript file printed into '%s'.\n\n", ps_file);
         else
         {
            if (short_name && *outputDir=='\0')
               sprintf (msg, "PostScript file printed into '%s'.", rest);
            else
               sprintf (msg, "PostScript file printed into '%s'.", ps_file);
            Msg (msg);
         }
         break;
   }
#ifdef KEEP_WHEN_PRINT
   if (whereToPrint != PRINTER) unlink (tmpFile);
#else
   unlink (tmpFile);
#endif
}

void Dump (FileName)
   char	* FileName;
{
   int			len, obj_ext_len, sym_ext_len;
   char			obj_ext_str[MAXSTRING+1], sym_ext_str[MAXSTRING+1];
   char			dummy_str[MAXSTRING+1];
   struct PageRec	* saved_cur_page;

   sprintf (obj_ext_str, ".%s", OBJ_FILE_EXT);
   obj_ext_len = strlen (obj_ext_str);
   sprintf (sym_ext_str, ".%s", SYM_FILE_EXT);
   sym_ext_len = strlen (sym_ext_str);

   if (whereToPrint == LATEX_FIG && pageLayoutMode == PAGE_TILE && (
         paperCol != 1 || paperRow != 1))
   {
      strcpy (dummy_str, "Can not print in EPS format in Tiled page mode.");
      if (PRTGIF)
         fprintf (stderr, "%s\n", dummy_str);
      else
         Dialog ("Can not print in EPS format in Tiled page mode.", NULL, NULL);
      return;
   }
   if (PRTGIF)
   {
      len = strlen (FileName);
      if ((len >= obj_ext_len &&
            strcmp (&FileName[len-obj_ext_len], obj_ext_str) == 0) ||
            (len >= sym_ext_len &&
            strcmp (&FileName[len-sym_ext_len], sym_ext_str) == 0))
         FileName[len-obj_ext_len] = '\0';
   }
   else
   {
      SetWatchCursor (drawWindow);
      SetWatchCursor (mainWindow);
   }
   totalBBoxValid = FALSE;
   if (whereToPrint == PRINTER || whereToPrint == PS_FILE)
   {
      printingFirstPageNum = 1;
      printingPageNum = 1;
      printingLastPageNum = lastPageNum;
      saved_cur_page = curPage;
      for (curPage = firstPage; curPage != NULL; curPage = curPage->next,
            printingPageNum++)
      {
         topObj = curPage->top;
         botObj = curPage->bot;
         GenDump (FileName);
      }
      curPage = saved_cur_page;
      topObj = curPage->top;
      botObj = curPage->bot;
   }
   else
   {
      printingFirstPageNum = 1;
      printingPageNum = 1;
      printingLastPageNum = 1;
      GenDump (FileName);
   }
   if (!PRTGIF)
   {
      SetDefaultCursor (mainWindow);
      ShowCursor ();
   }
}

void DumpOnePageInTileMode (row, col)
   int	row, col;
{
   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   totalBBoxValid = FALSE;
   printingFirstPageNum = printingPageNum = printingLastPageNum =
         row*paperCol+col+1;
   dumpOnePageInTileMode = TRUE;
   GenDump ("");
   dumpOnePageInTileMode = FALSE;

   SetDefaultCursor (mainWindow);
   ShowCursor ();
}

void DumpOnePageInStackMode ()
{
   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   totalBBoxValid = FALSE;
   printingFirstPageNum = printingPageNum = printingLastPageNum = curPageNum;
   dumpOnePageInStackMode = TRUE;
   GenDump ("");
   dumpOnePageInStackMode = FALSE;

   SetDefaultCursor (mainWindow);
   ShowCursor ();
}

void PrintWithCommand (FileName)
   char	* FileName;
{
   struct PageRec	* saved_cur_page;

   if (whereToPrint != PRINTER)
   {
      Msg ("PrintWithCmd only works when output device is the printer.");
      return;
   }
   Dialog ("Please Enter Print Command Name:",
         "( <CR>: accept, <ESC>: cancel )", printCommand);
   if (*printCommand == '\0') return;

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   printingPageNum = 1;
   printingLastPageNum = lastPageNum;
   saved_cur_page = curPage;
   totalBBoxValid = FALSE;
   for (curPage = firstPage; curPage != NULL; curPage = curPage->next,
         printingPageNum++)
   {
      topObj = curPage->top;
      botObj = curPage->bot;
      GenDump (FileName);
   }
   curPage = saved_cur_page;
   topObj = curPage->top;
   botObj = curPage->bot;

   SetDefaultCursor (mainWindow);
   ShowCursor ();
}

void PrintSelectedObjs ()
{
   char	dummy_str[MAXSTRING+1];
   struct SelRec	* top_sel_ptr, * bot_sel_ptr, * sel_ptr, * next_sel;
   struct ObjRec	* saved_top_obj, * saved_bot_obj, * obj_ptr;

   if (topSel == NULL)
   {
      Dialog ("No objects selected!  Nothing printed!",
            "( <CR> or <ESC> to continue )", dummy_str );
      return;
   }
   PushPageInfo ();
   saved_top_obj = topObj;
   saved_bot_obj = botObj;

   JustDupSelObj (&top_sel_ptr, &bot_sel_ptr);
   curPage->top = topObj = top_sel_ptr->obj;
   curPage->bot = botObj = bot_sel_ptr->obj;
   for (sel_ptr=topSel, obj_ptr=topObj; obj_ptr!=NULL;
         sel_ptr=sel_ptr->next, obj_ptr=obj_ptr->next)
   {
      CopyObjId (sel_ptr->obj, obj_ptr);
      CopyObjLocks (sel_ptr->obj, obj_ptr);
   }

   if (pageLayoutMode == PAGE_STACK)
      DumpOnePageInStackMode ();
   else
   {
      printingFirstPageNum = 1;
      printingPageNum = 1;
      printingLastPageNum = 1;
      totalBBoxValid = FALSE;
      GenDump ("");
   }
   
   DelAllObj ();
   for (sel_ptr = top_sel_ptr; sel_ptr != NULL; sel_ptr = next_sel)
   {
      next_sel = sel_ptr->next;
      cfree (sel_ptr);
   }
   PopPageInfo ();
   curPage->top = topObj = saved_top_obj;
   curPage->bot = botObj = saved_bot_obj;
}

void SetPrintReduction ()
{
   int	value;
   char	buf[MAXSTRING+1], msg[MAXSTRING+1];

   Dialog ("Please specify percent reduction (<100) or enlargement (>100):",
         "( <CR>: accept, <ESC>: cancel )", buf);
   if (*buf == '\0') return;

   value = atoi (buf);
   if (value <= 0)
   {
      sprintf (msg, "Invalid reduction '%s'.", buf);
      Msg (msg);
      return;
   }
   printMag = value;
   if (UpdPageStyle (pageStyle))
   {
      UpdDrawWinBBox ();
      AdjSplineVs ();
      ClearAndRedrawDrawWindow ();
   }
   RedrawScrollBars ();
   RedrawRulers ();
   RedrawTitleWindow ();
   SetFileModified (TRUE);
   if (printMag <= 100)
      sprintf (msg, "New reduction is %1d%%.", printMag);
   else
      sprintf (msg, "New enlargement is %1d%%.", printMag);
   Msg (msg);
}

void NewProc ()
{
   while (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before clear? [ync](y)",
            CONFIRM_YES))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: TieLooseEnds (); SetFileModified (FALSE); break;
         case CONFIRM_CANCEL: return;
      }
   }
   CleanUpComments ();
   CleanUpCmds ();
   CleanUpDrawingWindow ();
   ClearFileInfo ();
   ClearAndRedrawDrawWindow ();
   Msg ("Editing no file.");
   objId = 0;
   RedrawTitleWindow ();
   lastPageNum = 1;
   InitPage ();
   ShowPage ();
}

void OpenProc ()
{
   char	file_name[MAXPATHLENGTH+1];
   int	do_not_save = FALSE;

   while (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before open? [ync](y)",
            CONFIRM_YES))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: do_not_save = TRUE; SetFileModified (FALSE); break;
         case CONFIRM_CANCEL: return;
      }
   }
   if (SelectFileName ("Please select a file to OPEN ...",file_name) != INVALID)
   {
      char	ext_str[80];
      int	ext_len, len, obj_file=TRUE;

      sprintf (ext_str, ".%s", OBJ_FILE_EXT);
      ext_len = strlen (ext_str);
      len = strlen (file_name);

      if (len < ext_len ||
            strcmp (&file_name[len-ext_len], ext_str) != 0)
      {
         sprintf (ext_str, ".%s", SYM_FILE_EXT);
         ext_len = strlen (ext_str);

         if (len >= ext_len &&
               strcmp (&file_name[len-ext_len], ext_str) == 0)
            obj_file = FALSE;
      }
      if (!LoadFile (file_name, obj_file) && do_not_save)
         SetFileModified (TRUE);
   }
   else if (do_not_save)
      SetFileModified (TRUE);

   SetDefaultCursor (mainWindow);
   ShowCursor ();
}

char * fileMenuStr[] =
      { "New            ^n",
        "Open           ^o",
        "Save           ^s",
        "SaveNew       ^#s",
        "Import         #p",
        "ImportXBitmap ^#.",
        "ImportXPixmap ^#,",
        "ImportEPSFile  #(",
        "ChangeDomain   ^c",
        "Print          ^p",
        "PrintWithCmd   ^-",
        "PrintSelectedObjs",
        "PrintOnePage     ",
        "InputPolyPts     ",
        "InputPolygonPts  ",
        "Solve          #s",
        "Simulate       #y",
        "Probe          #b",
        "Animate        ^z",
        "Escape         #x",
        "SaveSelectedAs #~",
        "SaveSymInLibrary ",
        "Quit           ^q"
      };

int QuitProc ()
{
   int	do_not_save = FALSE;

   while (fileModified)
   {
      switch (YesNoCancel ("File modified, save file before quit? [ync](y)",
            CONFIRM_YES))
      {
         case CONFIRM_YES: SaveFile (); break;
         case CONFIRM_NO: do_not_save = TRUE; SetFileModified (FALSE); break;
         case CONFIRM_CANCEL: return (INVALID);
      }
   }
   if (AncesterModified ())
   {
      switch (YesNoCancel ("Ancester file modified, still quitting? [ync](y)",
            CONFIRM_YES))
      {
         case CONFIRM_YES: return (FILE_QUIT);
         case CONFIRM_NO:
            if (do_not_save) SetFileModified (TRUE);
            return (INVALID);
         case CONFIRM_CANCEL:
            if (do_not_save) SetFileModified (TRUE);
            return (INVALID);
      }
   }
   return (FILE_QUIT);
}

int SolveProc ()
{
   if (!saveTmpOnReturn) return (FILE_SOLVE);

   switch (SaveTmpFile ("tmpmodel"))
   {
      case OBJ_FILE_SAVED: return (FILE_SOLVE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
   return (INVALID);
}

int SimulateProc ()
{
   if (!saveTmpOnReturn) return (FILE_SIMULATE);

   switch (SaveTmpFile ("tmpmodel"))
   {
      case OBJ_FILE_SAVED: return (FILE_SIMULATE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
   return (INVALID);
}

int ProbeProc ()
{
   if (!saveTmpOnReturn) return (FILE_PROBE);

   switch (SaveTmpFile ("tmpmodel"))
   {
      case OBJ_FILE_SAVED: return (FILE_PROBE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
   return (INVALID);
}

int AnimateProc ()
{
   if (!saveTmpOnReturn) return (FILE_ANIMATE);

   switch (SaveTmpFile ("tmpmodel"))
   {
      case OBJ_FILE_SAVED: return (FILE_ANIMATE);
      case SYM_FILE_SAVED: return (INVALID);
      case INVALID: return (INVALID);
   }
   return (INVALID);
}

int EscapeProc ()
{
   return (FILE_ESCAPE);
}

int FileSubMenu (index)
   int	index;
{
   switch (index)
   {
      case FILE_NEW: NewProc (); break;
      case FILE_OPEN: OpenProc (); break;
      case FILE_SAVE: SaveFile (); break;
      case FILE_SAVENEW: SaveNewFile (FALSE); break;
      case FILE_IMPORT: ImportFile (); break;
      case FILE_IMPORTXBM: ImportXBitmapFile (); break;
      case FILE_IMPORTXPM: ImportXPixmapFile (); break;
      case FILE_IMPORTEPS: ImportEPSFile (); break;
      case FILE_DOMAIN: ChangeDomain (); break;
      case FILE_DUMP: Dump (""); break;
      case FILE_USR_DUMP: PrintWithCommand (""); break;
      case FILE_DUMPSELECTED: PrintSelectedObjs (); break;
      case FILE_PRINTONE: PrintOnePage (); break;
      case FILE_INPUT_POLY: InputPolyPts (); break;
      case FILE_INPUT_POLYGON: InputPolygonPts (); break;
      case FILE_SOLVE: return (SolveProc ());
      case FILE_SIMULATE: return (SimulateProc ());
      case FILE_PROBE: return (ProbeProc ());
      case FILE_ANIMATE: return (AnimateProc ());
      case FILE_ESCAPE: return (EscapeProc ());
      case FILE_SAVESELAS: SaveNewFile (TRUE); break;
      case FILE_SAVESYMINLIB: SaveSymInLibrary (); break;
      case FILE_QUIT: return (QuitProc ());
   }
   return (INVALID);
}

int FileMenu (X, Y)
   int	X, Y;
{
   register int index;
   int		* fore_colors, * valid, * init_rv;

   DefaultColorArrays (FILEMENUENTRIES, &fore_colors, &valid, &init_rv);
   activeMenu = MENU_FILE;
   index = TextMenuLoop (X, Y, fileMenuStr, FILEMENUENTRIES, fore_colors,
         valid, init_rv, SINGLECOLOR);

   if (index != INVALID) return (FileSubMenu (index));
   return (INVALID);
}

void CleanUpFiles ()
{
   CleanUpComments ();
   ClearFileInfo ();
   fileModified = FALSE;
}
